import os

os.environ["RUNNING_TEST"] = 'True'
import unittest
from tests.helpers import AppWindowTestCase
from pathlib import Path

from finalcif.appwindow import AppWindow

data = Path('tests')


class MyTestCase(AppWindowTestCase):

    def setUp(self) -> None:
        self.testcif = data / 'examples/1979688.cif'
        self.authorexport_file = data / 'examples/testexport_author.cif'
        self.testimport_author = data / 'other_templates/AATest_Author.cif'
        self.app = AppWindow(self.testcif)
        self.author = {'address': 'address', 'footnote': 'footnote', 'email': 'email',
                       'name'   : 'name', 'orcid': 'orcid', 'phone': 'phone', 'contact': True}
        self.app.ui.authorEditTabWidget.setCurrentIndex(0)
        self._delete_test_author()

    def tearDown(self) -> None:
        self.authorexport_file.unlink(missing_ok=True)
        self._delete_test_author()
        self.app.close()
        super().tearDown()

    def _import_testauthor(self):
        # To be used in other tests
        self.app.authors.import_author(str(self.testimport_author))
        self.app.ui.LoopTemplatesListWidget.setCurrentRow(0)

    def test_selected_loop_without_selection(self):
        self.assertEqual('', self.app.authors.get_selected_loop_name())

    def test_selected_loop_with_selection(self):
        self.assertEqual('', self.app.authors.get_selected_loop_name())
        self._import_testauthor()
        self.assertEqual('AATest Author', self.app.authors.get_selected_loop_name())

    def _delete_test_author(self, name='AATest Author'):
        self._select_row_of_author(name)
        self.app.ui.DeleteLoopAuthorTemplateButton.click()
        self.assertNotEqual('AATest Author', self.app.authors.get_selected_loop_name())

    def _select_row_of_author(self, name):
        listWidget = self.app.ui.LoopTemplatesListWidget
        for row in range(listWidget.count()):
            item = listWidget.item(row)
            if item is not None and item.text().startswith(name):
                listWidget.setCurrentRow(row)
                break

    def test_export_selected_author(self):
        self.authorexport_file.unlink(missing_ok=True)
        self._import_testauthor()
        self._select_row_of_author('AATest Author')
        self.app.authors.export_author_template(str(self.authorexport_file))
        self.assertEqual(True, self.authorexport_file.exists())
        self.assertEqual(self.testimport_author.read_text(), self.authorexport_file.read_text())

    def test_export_selected_author_cif(self):
        self.app.ui.authorEditTabWidget.setCurrentIndex(1)
        self.authorexport_file.unlink(missing_ok=True)
        self._import_testauthor()
        self._select_row_of_author('AATest Author')
        self.app.authors.export_author_template(str(self.authorexport_file))
        self.assertEqual(True, self.authorexport_file.exists())
        self.assertEqual(self.testimport_author.read_text(), self.authorexport_file.read_text())

    def test_set_name(self):
        self.app.ui.FullNameLineEdit.setText('test')
        self.assertEqual('test', self.app.authors.get_author_info().name)
        self.app.ui.FullNameLineEdit.clear()
        self.app.ui.authorEditTabWidget.setCurrentIndex(1)
        self.app.ui.FullNameLineEdit_cif.setText('test2')
        self.assertEqual('test2', self.app.authors.get_author_info().name)

    def test_set_contact_author(self):
        self.assertEqual(False, self.app.ui.ContactAuthorCheckBox.isChecked())
        self.app.ui.ContactAuthorCheckBox.setChecked(True)
        self.assertEqual(True, self.app.authors.get_author_info().contact_author)
        self.app.ui.authorEditTabWidget.setCurrentIndex(1)
        self.assertEqual(False, self.app.ui.ContactAuthorCheckBox_cif.isChecked())
        self.app.ui.ContactAuthorCheckBox_cif.setChecked(True)
        self.assertEqual(True, self.app.authors.get_author_info().contact_author)

    def test_set_address(self):
        self.app.ui.AddressTextedit.setText('Eine Adresse 1')
        self.assertEqual("'Eine Adresse 1'", self.app.authors.get_author_info().address)
        self.app.ui.AddressTextedit.clear()
        self.app.ui.authorEditTabWidget.setCurrentIndex(1)
        self.app.ui.AddressTextedit_cif.setText('Eine Adresse 2')
        self.assertEqual("'Eine Adresse 2'", self.app.authors.get_author_info().address)

    def test_set_footnote(self):
        self.app.ui.FootNoteLineEdit.setText('notex')
        self.assertEqual('notex', self.app.authors.get_author_info().footnote)

    def test_set_email(self):
        self.app.ui.EMailLineEdit.setText('test@foo.de')
        self.assertEqual('test@foo.de', self.app.authors.get_author_info().email)
        self.app.ui.authorEditTabWidget.setCurrentIndex(1)
        self.app.ui.EMailLineEdit.clear()
        self.app.ui.EMailLineEdit_cif.setText('test@foo.de')
        self.assertEqual('test@foo.de', self.app.authors.get_author_info().email)

    def test_set_orcid(self):
        self.app.ui.ORCIDLineEdit.setText('12345a')
        self.assertEqual('12345a', self.app.authors.get_author_info().orcid)

    def test_set_phone(self):
        self.app.ui.PhoneLineEdit.setText('12345a')
        self.assertEqual('12345a', self.app.authors.get_author_info().phone)
        self.app.ui.PhoneLineEdit.clear()
        self.app.ui.authorEditTabWidget.setCurrentIndex(1)
        self.app.ui.PhoneLineEdit_cif.setText('12345b')
        self.assertEqual('12345b', self.app.authors.get_author_info().phone)

    def test_set_foo(self):
        self.app.ui.IUCRIDLineEdit.setText('123-45a')
        self.assertEqual('123-45a', self.app.authors.get_author_info().iucr_id)

    def test_set_author_info(self):
        self.app.authors.set_author_info(self.author)
        self.assertEqual('name', self.app.ui.FullNameLineEdit.text())
        self.assertEqual('address', self.app.ui.AddressTextedit.toPlainText())
        self.assertEqual('email', self.app.ui.EMailLineEdit.text())
        self.assertEqual('orcid', self.app.ui.ORCIDLineEdit.text())
        self.assertEqual('footnote', self.app.ui.FootNoteLineEdit.text())
        self.assertEqual('phone', self.app.ui.PhoneLineEdit.text())
        self.assertEqual(True, self.app.ui.ContactAuthorCheckBox.isChecked())

    def test_set_author_info_and_clear(self):
        self.app.authors.set_author_info(self.author)
        self.app.authors.clear_fields()
        self.assertEqual('', self.app.ui.FullNameLineEdit.text())
        self.assertEqual('', self.app.ui.AddressTextedit.toPlainText())
        self.assertEqual('', self.app.ui.EMailLineEdit.text())
        self.assertEqual('', self.app.ui.ORCIDLineEdit.text())
        self.assertEqual('', self.app.ui.FootNoteLineEdit.text())
        self.assertEqual('', self.app.ui.PhoneLineEdit.text())
        self.assertEqual(False, self.app.ui.ContactAuthorCheckBox.isChecked())

    def test_import_multiple_authors(self):
        multi_author_cif = data / 'examples/multi_author_test.cif'
        self.app.authors.import_author(str(multi_author_cif))

        self._select_row_of_author('Contact Author')
        self.app.authors.load_selected_loop()
        self.assertEqual('Contact Author (contact author)', self.app.authors.get_selected_loop_name())
        self.assertEqual('single@contact.de', self.app.authors.get_author_info().email)
        self.assertEqual(True, self.app.authors.get_author_info().contact_author)
        self._delete_test_author('Contact Author (contact author)')

        self._select_row_of_author('Author One')
        self.app.authors.load_selected_loop()
        self.assertEqual('Author One', self.app.authors.get_selected_loop_name())
        self.assertEqual('one@test.de', self.app.authors.get_author_info().email)
        self.assertEqual(False, self.app.authors.get_author_info().contact_author)
        self._delete_test_author('Author One')

        self._select_row_of_author('Author Two')
        self.app.authors.load_selected_loop()
        self.assertEqual('Author Two', self.app.authors.get_selected_loop_name())
        self.assertEqual("''", self.app.authors.get_author_info().email)
        self.assertEqual(False, self.app.authors.get_author_info().contact_author)
        self._delete_test_author('Author Two')

if __name__ == '__main__':
    unittest.main()




import sys
import os.path
from pathlib import Path
from pprint import pprint

from PyQt5.QtCore import QCoreApplication, QLibraryInfo, QProcess, QProcessEnvironment


def main():
    app = QCoreApplication([])
    base_dir = Path(__file__).parent.parent
    plugins_dir = os.path.join(base_dir, "finalcif/gui/customWidgets")
    lib_dir = os.path.join(base_dir)

    env = QProcessEnvironment.systemEnvironment()
    #print(plugins_dir)
    env.insert("PYQTDESIGNERPATH", plugins_dir)
    #env.insert("QT_DEBUG_PLUGINS", '1')
    print('Designerpath:', os.environ.get('PYQTDESIGNERPATH'))
    env.insert("PYTHONPATH", '/Users/daniel/Documents/GitHub/FinalCif')
    #pprint([x for x in os.environ.items()])
    print(os.environ.get('PYTHONPATH'))

    # Start Designer.
    designer = QProcess()
    designer.setProcessEnvironment(env)

    designer_bin = QLibraryInfo.location(QLibraryInfo.BinariesPath)

    if sys.platform == "darwin":
        #designer_bin = '/usr/local/Cellar/qt/6.5.0/bin/Designer'
        designer_bin = '/usr/local/Cellar/qt@5/5.15.8_3/libexec/Designer.app/Contents/MacOS/Designer'
    else:
        designer_bin += "/designer"

    designer.start(designer_bin)
    designer.waitForFinished(-1)

    sys.exit(designer.exitCode())


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'new_key_dialog_ui.ui'
##
##
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QAbstractItemView, QApplication, QHBoxLayout, QLabel,
    QLineEdit, QListWidget, QListWidgetItem, QMainWindow,
    QPushButton, QSizePolicy, QVBoxLayout, QWidget)

class Ui_AddKeyWindow(object):
    def setupUi(self, AddKeyWindow):
        if not AddKeyWindow.objectName():
            AddKeyWindow.setObjectName(u"AddKeyWindow")
        AddKeyWindow.resize(528, 445)
        self.centralwidget = QWidget(AddKeyWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.searchLabel = QLabel(self.centralwidget)
        self.searchLabel.setObjectName(u"searchLabel")

        self.horizontalLayout.addWidget(self.searchLabel)

        self.searchLineEdit = QLineEdit(self.centralwidget)
        self.searchLineEdit.setObjectName(u"searchLineEdit")

        self.horizontalLayout.addWidget(self.searchLineEdit)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.keysListWidget = QListWidget(self.centralwidget)
        self.keysListWidget.setObjectName(u"keysListWidget")
        self.keysListWidget.setProperty(u"showDropIndicator", False)
        self.keysListWidget.setSelectionMode(QAbstractItemView.MultiSelection)
        self.keysListWidget.setSelectionBehavior(QAbstractItemView.SelectItems)

        self.verticalLayout.addWidget(self.keysListWidget)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.addKeyPushButton = QPushButton(self.centralwidget)
        self.addKeyPushButton.setObjectName(u"addKeyPushButton")

        self.horizontalLayout_2.addWidget(self.addKeyPushButton)

        self.cancelPushButton = QPushButton(self.centralwidget)
        self.cancelPushButton.setObjectName(u"cancelPushButton")

        self.horizontalLayout_2.addWidget(self.cancelPushButton)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        AddKeyWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(AddKeyWindow)
    # setupUi

    def retranslateUi(self, AddKeyWindow):
        AddKeyWindow.setWindowTitle(QCoreApplication.translate("AddKeyWindow", u"Add CIF Keyword", None))
        self.searchLabel.setText(QCoreApplication.translate("AddKeyWindow", u"Search", None))
        self.addKeyPushButton.setText(QCoreApplication.translate("AddKeyWindow", u"Add Key(s)", None))
        self.cancelPushButton.setText(QCoreApplication.translate("AddKeyWindow", u"Close", None))
    # retranslateUi




from __future__ import annotations

import threading
import time
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING

from qtpy import QtCore
from qtpy.QtWidgets import QListWidgetItem, QTableWidget, QStackedWidget, QLabel
from gemmi import cif

from finalcif.cif.text import retranslate_delimiter, utf8_to_str
from finalcif.equip_property.tools import read_document_from_cif_file
from finalcif.gui.dialogs import cif_file_open_dialog, show_general_warning, cif_file_save_dialog
from finalcif.gui.plaintextedit import PlainTextEditTemplate
from finalcif.tools import misc
from finalcif.tools.settings import FinalCifSettings

if TYPE_CHECKING:
    from finalcif.appwindow import AppWindow

class Properties(QtCore.QObject):
    def __init__(self, parent: AppWindow, settings: FinalCifSettings):
        super().__init__(parent=parent)
        self.app = parent
        self.settings = settings
        if parent:
            self.signals_and_slots()
            self.app.ui.PropertiesTemplatesStackedWidget.setCurrentIndex(0)
            self.app.ui.PropertiesEditTableWidget.verticalHeader().hide()
            self.store_predefined_templates()
            self.show_properties()
            self.lb = QLabel(self.app)

    def signals_and_slots(self):
        ## properties
        self.app.ui.EditPropertyTemplateButton.clicked.connect(self.edit_property_template)
        self.app.ui.SavePropertiesButton.clicked.connect(self.save_property_template)
        self.app.ui.CancelPropertiesButton.clicked.connect(self.cancel_property_template)
        self.app.ui.DeletePropertiesButton.clicked.connect(self.delete_property)
        ## properties
        self.app.ui.PropertiesEditTableWidget.itemSelectionChanged.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.cellPressed.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.itemEntered.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.cellChanged.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.currentItemChanged.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.itemActivated.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.itemPressed.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.itemClicked.connect(self.add_property_row_if_needed)
        self.app.ui.PropertiesEditTableWidget.itemChanged.connect(self.add_property_row_if_needed)
        self.app.ui.NewPropertyTemplateButton.clicked.connect(self.new_property)
        self.app.ui.ImportPropertyTemplateButton.clicked.connect(self.import_property_from_file)
        self.app.ui.ExportPropertyButton.clicked.connect(self.export_property_template)
        self.app.ui.cifKeywordLineEdit.textChanged.connect(self.check_for_duplicates)

    def check_for_duplicates(self) -> None:
        key = self.app.ui.cifKeywordLineEdit.text()
        props = self.export_raw_data()
        keys = [x['cif_key'] for x in props]
        names = [x['name'] for x in props]
        if key in keys and self.selected_template_name() not in names:
            self.app.ui.SavePropertiesButton.setDisabled(True)
            self.lb.setWindowFlags(QtCore.Qt.WindowType.ToolTip)
            self.lb.setText(f'key {key} already exists')
            self.lb.move(self.app.ui.cifKeywordLineEdit.mapToGlobal(QtCore.QPoint(15, 25)))
            self.lb.show()
            QtCore.QTimer().singleShot(5000, self.hide_label)
        else:
            self.app.ui.SavePropertiesButton.setEnabled(True)
            self.lb.hide()

    def hide_label(self) -> None:
        self.lb.hide()

    def show_properties(self) -> None:
        """
        Display saved items in the properties lists.
        """
        self.app.ui.PropertiesTemplatesListWidget.clear()
        property_list = self.settings.get_properties_list()
        for pr in property_list:
            if pr:
                item = QListWidgetItem(pr)
                self.app.ui.PropertiesTemplatesListWidget.addItem(item)

    def new_property(self) -> None:
        item = QListWidgetItem('')
        self.app.ui.PropertiesTemplatesListWidget.addItem(item)
        self.app.ui.PropertiesTemplatesListWidget.setCurrentItem(item)
        item.setFlags(QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable)
        self.app.ui.PropertiesTemplatesListWidget.editItem(item)
        self.app.ui.cifKeywordLineEdit.clear()

    def add_property_row_if_needed(self) -> None:
        """
        Adds an empty row at the bottom of either the PropertiesEditTableWidget.
        """
        table = self.app.ui.PropertiesEditTableWidget
        rowcount = table.rowCount()
        cont = 0
        for row in range(rowcount):
            key = ''
            with suppress(AttributeError, TypeError):
                key = table.item(row, 0).text()
            with suppress(AttributeError, TypeError):
                key = table.cellWidget(row, 0).getText()
            if key:  # don't count empty key rows
                cont += 1
        diff = rowcount - cont
        if diff < 4:
            table.insertRow(rowcount)

    # The properties templates:

    def delete_property(self) -> None:
        # First delete the list entries
        index = self.app.ui.PropertiesTemplatesListWidget.currentIndex()
        property_name = index.data()
        self.settings.delete_template('property', property_name)
        # now make it invisible:
        self.app.ui.PropertiesTemplatesListWidget.takeItem(index.row())
        self.cancel_property_template()
        # I do these both to clear the list:
        self.store_predefined_templates()
        self.show_properties()
        self.app.refresh_combo_boxes()

    def edit_property_template(self) -> None:
        """
        Edit the Property table.
        """
        # make sure the current item doesn't get lost:
        it = self.app.ui.PropertiesTemplatesListWidget.currentItem()
        self.app.ui.PropertiesTemplatesListWidget.setCurrentItem(None)
        self.app.ui.PropertiesTemplatesListWidget.setCurrentItem(it)
        self.app.ui.CancelEquipmentButton.click()
        self.load_property_from_settings()

    def save_property_template(self) -> None:
        table = self.app.ui.PropertiesEditTableWidget
        stackedwidget = self.app.ui.PropertiesTemplatesStackedWidget
        keyword = self.app.ui.cifKeywordLineEdit.text()
        self.save_property(table, stackedwidget, keyword)
        self.app.refresh_combo_boxes()

    def store_predefined_templates(self) -> None:
        property_list = self.settings.get_properties_list() or []
        for item in misc.predefined_property_templates:
            if item['name'] not in property_list:
                self.settings.save_settings_list('property', item['name'], item['values'])

    def export_property_template(self, filename: str = '') -> None:
        """
        Exports the currently selected property entry to a file.
        """
        if not self.selected_template_name():
            return
        prop_data = self.settings.load_settings_list('property', self.selected_template_name())
        table_data = []
        cif_key = ''
        if prop_data:
            cif_key = prop_data[0]
            with suppress(Exception):
                table_data = prop_data[1]
        if not cif_key:
            return
        doc = cif.Document()
        blockname = '__'.join(self.selected_template_name().split())
        block = doc.add_new_block(blockname)
        try:
            loop = block.init_loop(cif_key, [''])
        except RuntimeError:
            # Not a valid loop key
            show_general_warning(self.app, f'"{cif_key}" is not a valid cif keyword.')
            return
        for value in table_data:
            if value:
                loop.add_row([cif.quote(utf8_to_str(value))])
        if not filename:
            filename = cif_file_save_dialog(blockname.replace('__', '_') + '.cif')
        if not filename.strip():
            return
        try:
            doc.write_file(filename, style=cif.Style.Indent35)
        except PermissionError:
            if Path(filename).is_dir():
                return
            show_general_warning(self.app, f'No permission to write file to {Path(filename).resolve()}')

    def selected_template_name(self) -> str:
        return self.app.ui.PropertiesTemplatesListWidget.currentIndex().data()

    def import_property_from_file(self, filename: str = '') -> None:
        """
        Imports a cif file as entry of the property templates list.
        """
        if not filename:
            filename = cif_file_open_dialog(filter="CIF file (*.cif)")
        if not filename:
            return
        doc = read_document_from_cif_file(filename)
        if not doc:
            return
        for block in doc:
            self._import_block(block)
        self.show_properties()

    def _import_block(self, block: cif.Block) -> None:
        property_list = self.settings.load_property_list()
        if not property_list:
            property_list = ['']
        template_list = []
        loop_column_name = ''
        for i in block:
            if i.loop is not None:
                if len(i.loop.tags) > 0:
                    loop_column_name = i.loop.tags[0]
                for n in range(i.loop.length()):
                    value = i.loop[n, 0]
                    template_list.append(retranslate_delimiter(cif.as_string(value).strip("\n\r ;")))
        block_name = block.name.replace('__', ' ')
        # This is the list shown in the Main menu:
        property_list.append(block_name)
        table = self.app.ui.PropertiesEditTableWidget
        table.setRowCount(0)
        self.app.ui.cifKeywordLineEdit.setText(loop_column_name)
        newlist = [x for x in list(set(property_list)) if x]
        newlist.sort()
        template_list.insert(0, '')
        template_list = list(set(template_list))
        # save as dictionary for properties to have "_cif_key : itemlist"
        # for a table item as dropdown menu in the main table.
        table_data = [loop_column_name, template_list]
        self.settings.save_settings_list(property='property', name=block_name, items=table_data)

    def load_property_from_settings(self) -> None:
        """
        Load/Edit the value list of a property entry.
        """
        if not self.selected_template_name():
            # nothing selected
            return
        table = self.app.ui.PropertiesEditTableWidget
        table.blockSignals(True)
        table.clearContents()
        table.setRowCount(0)
        table_data = self.settings.load_settings_list('property', self.selected_template_name())
        if table_data:
            cif_key = table_data[0]
            with suppress(Exception):
                table_data = table_data[1]
            self.app.ui.cifKeywordLineEdit.setText(cif_key)
        if not table_data:
            table_data = ['', '', '']
        for value in table_data:
            try:
                self.add_propeties_row(table, retranslate_delimiter(str(value)))
            except TypeError:
                print('Bad value in property table')
                continue
        self.add_propeties_row(table, '')
        self.app.ui.PropertiesTemplatesStackedWidget.setCurrentIndex(1)
        table.blockSignals(False)
        # table.setWordWrap(False)
        table.resizeRowsToContents()

    def export_raw_data(self) -> list[dict]:
        properties_list = []
        for property_name in self.settings.get_properties_list():
            if property_name:
                property_cif_key, property_data = self.settings.load_settings_list('property', item_name=property_name)
                properties_list.append({'name': property_name, 'cif_key': property_cif_key, 'data': property_data})
        return properties_list

    def import_raw_data(self, properties_list: list[dict]) -> None:
        for property in properties_list:
            self.settings.save_settings_list(property='property', name=property.get("name"),
                                             items=[property.get('cif_key'), property.get('data')])
        self.show_properties()

    @staticmethod
    def add_propeties_row(table: QTableWidget, value: str = '') -> None:
        """
        Add a new row with a value to the Property table.
        """
        # Create a empty row at bottom of table
        row_num = table.rowCount()
        table.insertRow(row_num)
        key_item = PlainTextEditTemplate(parent=table)
        key_item.textChanged.connect(lambda: table.resizeRowsToContents())
        key_item.setPlainText(value)
        table.setCellWidget(row_num, 0, key_item)

    def save_property(self, table: QTableWidget,
                      stackwidget: QStackedWidget,
                      keyword: str = '') -> None:
        """
        Saves the currently selected Property template to the config file.
        """
        # Set None Item to prevent loss of the currently edited item:
        # The current item is closed and thus saved.
        table.setCurrentItem(None)
        table_data = []
        ncolumns = table.rowCount()
        for rownum in range(ncolumns):
            try:
                # only one column!
                value = table.cellWidget(rownum, 0).getText()
            except AttributeError:
                value = ''
            if value:
                table_data.append(value)
        # make sure to have always a blank item first:
        table_data.insert(0, '')
        if keyword:
            # save as list for properties to have [_cif_key, itemlist]
            # for a table item as dropdown menu in the main table.
            table_data = [keyword, table_data]
        self.settings.save_settings_list(property='property', name=self.selected_template_name(), items=table_data)
        stackwidget.setCurrentIndex(0)
        print('saved')

    def cancel_property_template(self) -> None:
        """
        Cancel editing of the current template.
        """
        table = self.app.ui.PropertiesEditTableWidget
        table.clearContents()
        table.setRowCount(0)
        self.app.ui.PropertiesTemplatesStackedWidget.setCurrentIndex(0)


if __name__ == '__main__':
    l = Properties(None, FinalCifSettings())
    for line in l.export_raw_data():
        print(f"{line}")
    print(l.export_raw_data())



import sys

import requests
from qtpy.QtCore import Signal, QObject

from finalcif import VERSION


class MyDownloader(QObject):
    progress = Signal(str)
    failed = Signal(int)
    finished = Signal()
    loaded = Signal(bytes)

    def __init__(self, url, parent=None):
        super().__init__(parent)
        self.url = url

    def failed_to_load(self, txt: str) -> None:
        print('Failed to load {}', self.url)

    def download(self) -> None:
        # print('Downloading:', self.url)
        OS = sys.platform
        user_agent = f'FinalCif v{VERSION} ({OS})'
        headers = {
            'User-Agent': user_agent,
        }
        try:
            response = requests.get(self.url, stream=True, headers=headers, timeout=10)
        except requests.RequestException as e:
            print('Could not connect to download server', e)
            self.loaded.emit(b'')
            self.failed.emit(0)
            return
        if response.status_code != 200:
            # noinspection PyUnresolvedReferences
            self.failed.emit(response.status_code)
            # noinspection PyUnresolvedReferences
            self.finished.emit()
            return
        self.loaded.emit(response.content)
        self.finished.emit()


if __name__ == "__main__":
    from qtpy.QtWidgets import QWidget, QHBoxLayout, QLabel
    from qtpy.QtWidgets import QApplication
    import threading

    app = QApplication(sys.argv)
    w = QWidget()
    l = QHBoxLayout()
    label = QLabel()
    l.addWidget(label)
    w.setLayout(l)
    w.setMinimumWidth(100)
    w.setMinimumHeight(100)
    w.show()


    def foo(bar: bytes):
        label.setText(bar.decode('ascii'))


    worker = MyDownloader("https://dkratzert.de/files/finalcif/version.txt")
    worker.loaded.connect(foo)
    thread = threading.Thread(target=worker.download, daemon=True)
    thread.start()
    sys.exit(app.exec())



from __future__ import annotations

from qtpy import QtCore
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QTableWidgetItem

# Re-export Column from its canonical location so existing imports still work.
from finalcif.gui.cif_table_model import Column  # noqa: F401

white = QColor(255, 255, 255)
light_green = QColor(217, 255, 201)
light_blue = QColor(244, 244, 249)
blue = QColor(102, 150, 179)
yellow = QColor(250, 247, 150)  # #faf796
light_red = QColor(254, 191, 189)


class MyTableWidgetItem(QTableWidgetItem):

    def __init__(self, *args, **kwargs):
        # args and kwargs are essential here. Otherwise, the horizontal header text is missing!
        super().__init__(*args, **kwargs)

    def setUneditable(self) -> None:
        # noinspection PyTypeChecker
        self.setFlags(self.flags() ^ QtCore.Qt.ItemFlag.ItemIsEditable)
        # noinspection PyTypeChecker
        self.setFlags(self.flags() | QtCore.Qt.ItemFlag.ItemIsSelectable)


class CifOrderItem(MyTableWidgetItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setUneditable()

    def isEssential(self):
        return self.checkState() == Qt.Checked

    def setEssential(self, state: bool):
        if state is True:
            self.setCheckState(Qt.Checked)
        else:
            self.setCheckState(Qt.Unchecked)



import shutil
from pathlib import Path

from qtpy import QtWidgets, QtCore

from finalcif.cif.cif_file_io import CifContainer
from finalcif.gui.import_selector_ui import Ui_importSelectMainWindow
from finalcif.tools import misc
from finalcif.tools.misc import do_not_import_keys, do_not_import_from_stoe_cfx
from finalcif.tools.settings import FinalCifSettings


class ImportSelector(QtWidgets.QMainWindow):
    import_clicked = QtCore.Signal(list, list)

    def __init__(self, parent, import_cif: CifContainer, target_cif: CifContainer, settings: FinalCifSettings):
        super().__init__(parent)
        self.import_cif = import_cif
        self.target_cif = target_cif
        self.settings = settings
        self.ui = Ui_importSelectMainWindow()
        self.ui.setupUi(self)
        self.show()
        self.keys_to_import: int = 0
        self.loops_to_import: int = 0
        self.selected: int = 0
        self._connect_signals_and_slots()

    def _connect_signals_and_slots(self) -> None:
        self.ui.saveSelectionPushbutton.clicked.connect(self._save_selection)
        self.ui.selectOnlyNewPB.clicked.connect(self._select_only_new)
        self.ui.importSelectedPushbutton.clicked.connect(self.import_key_loop)

    def import_key_loop(self) -> None:
        self.import_clicked.emit(self.get_keys(include=True), self.get_loops(include=True))

    def show_import_window(self) -> None:
        row = 0
        excluded_kv, excluded_loops = self._get_excluded_items()
        for item in self.import_cif.block:
            if item.pair is not None:
                key, _ = item.pair
                self._add_checkbox(key, row, self.ui.importTable_keys, checked=key not in excluded_kv)
                self.keys_to_import += 1
            else:
                continue
            row += 1
        row = 0
        for item in self.import_cif.block:
            if item.loop is not None:
                first_key = item.loop.tags[0]
                self.loops_to_import += 1
                key = '\n'.join(list(item.loop.tags))
                self._add_checkbox(key, row, self.ui.importTable_loops, checked=first_key not in excluded_loops)
            else:
                continue
            row += 1
        self.ui.importTable_keys.horizontalHeader().setStretchLastSection(True)
        self.ui.importTable_loops.horizontalHeader().setStretchLastSection(True)
        self.ui.importTable_keys.resizeRowsToContents()
        self.ui.importTable_loops.resizeRowsToContents()
        self._set_label()

    def _save_selection(self) -> None:
        self.settings.save_key_value('do_not_import_keys', self.get_keys(include=False))
        self.settings.save_key_value('do_not_import_loops', self.get_loops(include=False))

    def _empty_saved_selection(self) -> None:
        self.settings.save_key_value('do_not_import_keys', [])
        self.settings.save_key_value('do_not_import_loops', [])

    def _get_excluded_items(self) -> tuple[list[str], list[list[str]]]:
        excluded_kv = misc.do_not_import_keys
        excluded_loops = misc.do_not_loop_import
        exclude_kv_from_settings = self.settings.load_value_of_key('do_not_import_keys')
        exclude_loops_from_settings = self.settings.load_value_of_key('do_not_import_loops')
        if exclude_kv_from_settings:
            excluded_kv = exclude_kv_from_settings
        if exclude_loops_from_settings:
            excluded_loops = exclude_loops_from_settings
        return excluded_kv, excluded_loops

    def _set_label(self) -> None:
        self.ui.importInfoLabel.setText(f"Importing\n{self.import_cif.fileobj}\n\n"
                                        f"The CIF to import contains {self.keys_to_import} keys "
                                        f"and {self.loops_to_import} loops from which "
                                        f"{len(self.get_keys(include=True)) + len(self.get_loops(include=True))} "
                                        f"are selected for import.")

    def _add_checkbox(self, text: str, row: int, col: QtWidgets.QTableWidget, checked: bool = False) -> None:
        if col.rowCount() <= row:
            col.insertRow(row)
        checkbox = QtWidgets.QCheckBox(col)
        checkbox.stateChanged.connect(self._set_label)
        checkbox.setText(text)
        col.setCellWidget(row, 0, checkbox)
        checkbox.setChecked(checked)

    def get_keys(self, include: bool) -> list[str]:
        keys = []
        rows = self.ui.importTable_keys.rowCount()
        for row in range(rows):
            widget: QtWidgets.QCheckBox = self.ui.importTable_keys.cellWidget(row, 0)
            if widget and widget.isChecked() == include:
                keys.append(widget.text())
        return keys

    def get_loops(self, include: bool) -> list[list[str]]:
        loops = []
        rows = self.ui.importTable_loops.rowCount()
        for row in range(rows):
            widget: QtWidgets.QCheckBox = self.ui.importTable_loops.cellWidget(row, 0)
            if widget and widget.isChecked() == include:
                loop: list[str] = widget.text().splitlines(keepends=False)
                loops.append(loop)
        return loops

    def do_not_import_this_key(self, key: str) -> bool:
        value = self.import_cif[key]
        if value == '?' or value.strip() == '':
            return True
        if key in do_not_import_keys:
            return True
        if key in do_not_import_from_stoe_cfx and '.cfx' in self.import_cif.fileobj.name:
            return True
        return False

    def _select_only_new(self) -> None:
        self.select_pairs()
        self.select_loops()

    def select_pairs(self):
        rows = self.ui.importTable_keys.rowCount()
        for row in range(rows):
            widget: QtWidgets.QCheckBox = self.ui.importTable_keys.cellWidget(row, 0)
            key = widget.text()
            if self.target_cif[key]:
                # Import only new key/values
                widget.setChecked(False)
                continue
            widget.setChecked(True)

    def select_loops(self):
        rows = self.ui.importTable_loops.rowCount()
        for row in range(rows):
            widget: QtWidgets.QCheckBox = self.ui.importTable_loops.cellWidget(row, 0)
            loop = widget.text()
            first_key = loop.splitlines(keepends=False)[0]
            if self.target_cif.block.find_loop(first_key):
                # Import only new key/values
                widget.setChecked(False)
                continue
            widget.setChecked(True)


if __name__ == "__main__":
    import sys

    app = QtWidgets.QApplication([])
    imp_cif = CifContainer('test-data/DK_Zucker2_0m.cif')
    shutil.copyfile('test-data/p21c.cif', 'test-data/p21c-copy.cif')
    targetcif = CifContainer('test-data/p21c-copy.cif')
    settings = FinalCifSettings()
    imp = ImportSelector(None, import_cif=imp_cif, target_cif=targetcif, settings=settings)
    imp.show_import_window()
    app.exec()
    Path('test-data/p21c-copy.cif').unlink(missing_ok=True)
    sys.exit()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'loop_creator_ui.ui'
##
##
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QApplication, QHBoxLayout, QLabel, QLineEdit,
    QListWidget, QListWidgetItem, QPushButton, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_LoopCreator(object):
    def setupUi(self, LoopCreator):
        if not LoopCreator.objectName():
            LoopCreator.setObjectName(u"LoopCreator")
        LoopCreator.resize(800, 600)
        self.horizontalLayout_2 = QHBoxLayout(LoopCreator)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.label_2 = QLabel(LoopCreator)
        self.label_2.setObjectName(u"label_2")

        self.verticalLayout_2.addWidget(self.label_2)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.label = QLabel(LoopCreator)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.searchLineEdit = QLineEdit(LoopCreator)
        self.searchLineEdit.setObjectName(u"searchLineEdit")

        self.horizontalLayout.addWidget(self.searchLineEdit)


        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.availableKeysListWidget = QListWidget(LoopCreator)
        self.availableKeysListWidget.setObjectName(u"availableKeysListWidget")

        self.verticalLayout_2.addWidget(self.availableKeysListWidget)


        self.horizontalLayout_2.addLayout(self.verticalLayout_2)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, -1, -1, -1)
        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_2)

        self.rightPushButton = QPushButton(LoopCreator)
        self.rightPushButton.setObjectName(u"rightPushButton")

        self.verticalLayout.addWidget(self.rightPushButton)

        self.leftPushButton = QPushButton(LoopCreator)
        self.leftPushButton.setObjectName(u"leftPushButton")

        self.verticalLayout.addWidget(self.leftPushButton)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)


        self.horizontalLayout_2.addLayout(self.verticalLayout)

        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.label_3 = QLabel(LoopCreator)
        self.label_3.setObjectName(u"label_3")

        self.verticalLayout_3.addWidget(self.label_3)

        self.label_4 = QLabel(LoopCreator)
        self.label_4.setObjectName(u"label_4")

        self.verticalLayout_3.addWidget(self.label_4)

        self.newLoopKeysListWidget = QListWidget(LoopCreator)
        self.newLoopKeysListWidget.setObjectName(u"newLoopKeysListWidget")

        self.verticalLayout_3.addWidget(self.newLoopKeysListWidget)

        self.saveLoopPushButton = QPushButton(LoopCreator)
        self.saveLoopPushButton.setObjectName(u"saveLoopPushButton")

        self.verticalLayout_3.addWidget(self.saveLoopPushButton)


        self.horizontalLayout_2.addLayout(self.verticalLayout_3)


        self.retranslateUi(LoopCreator)
    # setupUi

    def retranslateUi(self, LoopCreator):
        LoopCreator.setWindowTitle(QCoreApplication.translate("LoopCreator", u"LoopCreator", None))
        self.label_2.setText(QCoreApplication.translate("LoopCreator", u"<html><head/><body><p><span style=\" font-size:18pt;\">Available CIF keywords</span></p></body></html>", None))
        self.label.setText(QCoreApplication.translate("LoopCreator", u"Search a Key", None))
        self.rightPushButton.setText(QCoreApplication.translate("LoopCreator", u"-->", None))
        self.leftPushButton.setText(QCoreApplication.translate("LoopCreator", u"<--", None))
        self.label_3.setText(QCoreApplication.translate("LoopCreator", u"<html><head/><body><p><span style=\" font-size:18pt;\">New Loop Header</span></p></body></html>", None))
        self.label_4.setText(QCoreApplication.translate("LoopCreator", u"Add CIF keys in order to create a new loop", None))
        self.saveLoopPushButton.setText(QCoreApplication.translate("LoopCreator", u"Save new Loop", None))
    # retranslateUi




from __future__ import annotations

import enum
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import gemmi
from qtpy import QtWidgets, QtCore

from finalcif.cif import cif_order
from finalcif.ciforder.order_ui import Ui_CifOrderForm
from finalcif.gui import dialogs
from finalcif.gui.custom_classes import CifOrderItem
from finalcif.gui.dialogs import cif_file_save_dialog, show_general_warning
from finalcif.gui.new_key_dialog import NewKey

if TYPE_CHECKING:
    from finalcif.tools.settings import FinalCifSettings


class Column(enum.IntEnum):
    key = 0


class CifOrder(QtWidgets.QGroupBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_CifOrderForm()
        self.ui.setupUi(self)
        self.cif = None
        self.settings = None
        self.essential_keys = cif_order.essential_keys
        self.set_keys(cif_order.order)
        self.connect_signals_and_slots()

    def set_order_from_settings(self, settings: FinalCifSettings = None):
        self.settings = settings
        if settings is not None:
            order = settings.load_settings_list('cif_order', 'order')
            essentials = settings.load_settings_list('cif_order', 'essentials')
            self.essential_keys = essentials
            if order:
                self.essential_keys = essentials
                self.set_keys(order)
            else:
                self.essential_keys = cif_order.essential_keys
                self.set_keys(cif_order.order)

    def connect_signals_and_slots(self):
        self.ui.importCifPushButton.clicked.connect(self.import_cif)
        self.ui.restoreDefaultPushButton.clicked.connect(self.restore_default)
        self.ui.addKeyPushButton.clicked.connect(self.open_add_cif_key)
        self.ui.saveSettingPushButton.clicked.connect(self._save_setting)
        self.ui.deleteKeyPushButton.clicked.connect(self.delete_keys)
        self.ui.exportToCifPushButton.clicked.connect(self.export_cif)
        # Connect signals for moveUpPushButton
        self.ui.moveUpPushButton.pressed.connect(lambda: self.start_repeated_action(self.move_row_up))
        self.ui.moveUpPushButton.released.connect(self.stop_repeated_action)
        # Connect signals for moveDownPushButton
        self.ui.moveDownPushButton.pressed.connect(lambda: self.start_repeated_action(self.move_row_down))
        self.ui.moveDownPushButton.released.connect(self.stop_repeated_action)
        # Timer for repeated actions
        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(170)
        self.timer.timeout.connect(self.perform_repeated_action)
        self.current_action = None

    def start_repeated_action(self, action):
        """Start the timer with the given action."""
        self.current_action = action
        action()  # Perform the action immediately
        self.timer.start()

    def stop_repeated_action(self):
        """Stop the timer and reset the current action."""
        self.timer.stop()
        self.current_action = None

    def perform_repeated_action(self):
        """Perform the currently set action."""
        if self.current_action:
            self.current_action()

    def move_row_up(self):
        current_row = self.ui.cifOrderTableWidget.currentRow()
        if current_row > 0:
            self.swap_rows(current_row, current_row - 1)
            self.ui.cifOrderTableWidget.setCurrentCell(current_row - 1, Column.key)

    def move_row_down(self):
        table = self.ui.cifOrderTableWidget
        current_row = table.currentRow()
        if current_row < table.rowCount() - 1:
            self.swap_rows(current_row, current_row + 1)
            table.setCurrentCell(current_row + 1, Column.key)

    def swap_rows(self, row1, row2):
        table = self.ui.cifOrderTableWidget
        item1 = table.item(row1, Column.key)
        item2 = table.item(row2, Column.key)
        if item1 and item2:
            text1 = item1.text()
            text2 = item2.text()
            check_state1 = item1.checkState()
            check_state2 = item2.checkState()
            item1.setText(text2)
            item1.setCheckState(check_state2)
            item2.setText(text1)
            item2.setCheckState(check_state1)

    @property
    def order_keys(self):
        keys = []
        for row in range(self.ui.cifOrderTableWidget.rowCount()):
            key = self.ui.cifOrderTableWidget.item(row, Column.key).text()
            keys.append(key)
        return keys

    @property
    def order_essentials(self):
        essentials = []
        for row in range(self.ui.cifOrderTableWidget.rowCount()):
            item = self.ui.cifOrderTableWidget.item(row, Column.key)
            if item.checkState() == QtCore.Qt.CheckState.Checked:
                essentials.append(item.text())
        return essentials

    def set_keys_from_cif(self, cif_file) -> None:
        try:
            self.cif = gemmi.cif.read_file(cif_file)
            block = self.cif.sole_block()
        except Exception:
            show_general_warning(self, 'Unable to load file')
            return
        keys = [x.pair[0] for x in block if x.pair is not None]
        essentials = [x.pair[0] for x in block if x.pair is not None and x.pair[1] == 'True']
        self.set_essentials(essentials)
        self.set_keys(keys)

    def set_essentials(self, essentials):
        self.essential_keys = essentials

    def set_keys(self, order_keys: list[str] | None = None):
        self.ui.cifOrderTableWidget.setRowCount(0)
        row = 0
        for key_text in order_keys:
            if not key_text:
                continue
            self.ui.cifOrderTableWidget.insertRow(row)
            self.set_row_text(key_text, row)
            row += 1
        # self.ui.cifOrderTableWidget.resizeColumnsToContents()

    def set_row_text(self, key_text: str, row: int) -> None:
        item1 = CifOrderItem(key_text)
        item1.setFlags(item1.flags() | QtCore.Qt.ItemFlag.ItemIsUserCheckable | QtCore.Qt.ItemFlag.ItemIsEnabled)
        item1.setText(key_text)
        if key_text in self.essential_keys:
            item1.setEssential(True)
        else:
            item1.setEssential(False)
        self.ui.cifOrderTableWidget.setItem(row, Column.key, item1)

    def set_keys_from_settings(self, keys: list[str]):
        self.set_keys(keys)

    def import_cif(self):
        cif_file = dialogs.cif_file_open_dialog()
        if self._file_is_there(cif_file):
            self.set_keys_from_cif(cif_file)

    def _file_is_there(self, cif_file: str) -> bool:
        return cif_file is not None and Path(cif_file).is_file() and Path(cif_file).exists()

    def restore_default(self):
        self.essential_keys = cif_order.essential_keys
        self.settings.save_settings_list('cif_order', 'order', cif_order.order)
        self.settings.save_settings_list('cif_order', 'essentials', cif_order.essential_keys)
        self.set_keys(cif_order.order)

    def open_add_cif_key(self):
        new_key = NewKey(self)
        new_key.show()
        new_key.new_key_added.connect(self.add_key)

    def add_key(self, key: str) -> None:
        if key not in self.order_keys:
            self.ui.cifOrderTableWidget.insertRow(0)
            self.set_row_text(key, 0)

    def _save_setting(self) -> None:
        self.save_in_settings(order_keys=self.order_keys, order_essentials=self.order_essentials)

    def save_in_settings(self, order_keys: list[str], order_essentials: list[str]) -> None:
        self.settings.save_settings_list('cif_order', 'order', order_keys)
        self.settings.save_settings_list('cif_order', 'essentials', order_essentials)
        self.essential_keys = order_essentials

    def delete_keys(self):
        items = self.ui.cifOrderTableWidget.selectedItems()
        for item in items:
            self.ui.cifOrderTableWidget.removeRow(item.row())

    def export_cif(self):
        doc = gemmi.cif.Document()
        blockname = 'FinalCif_keys_order'
        block = doc.add_new_block(blockname)
        for key in self.order_keys:
            value = 'True' if key in self.order_essentials else 'False'
            block.set_pair(key, value)
        filename = cif_file_save_dialog(blockname.lower() + '.cif')
        if not filename.strip():
            return
        try:
            doc.write_file(filename, style=gemmi.cif.Style.Indent35)
        except PermissionError:
            if Path(filename).is_dir():
                return
            show_general_warning(self, f'No permission to write file to {Path(filename).resolve()}')


if __name__ == "__main__":
    from finalcif.tools.settings import FinalCifSettings

    app = QtWidgets.QApplication(sys.argv)
    # form = CifOrder(parent=None, cif_file=Path('test-data/1000007.cif').resolve())
    settings = FinalCifSettings()
    form = CifOrder(parent=None)
    form.set_order_from_settings(settings)
    form.show()
    form.raise_()
    sys.exit(app.exec())



import sys

from qtpy import QtGui, QtCore
from qtpy.QtCore import QSize, Signal
from qtpy.QtWidgets import QWidget, QHBoxLayout, QCheckBox, QListWidgetItem, QVBoxLayout, QLabel

from finalcif.gui import text_templates_ui
from finalcif.gui.spell_check_edit import SpellTextEdit


class TextEditItem(QWidget):
    """
    Text editor for large text inside of dropdown widgets.
    """
    _num = 1
    checkbox_clicked = Signal(str)

    def __init__(self, parent: QWidget):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        self.vlayout = QVBoxLayout()
        self.vlayout.setAlignment(QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.checkbox = QCheckBox()
        self.number_label = QLabel()
        self.vlayout.addWidget(self.number_label)
        self.textfield = SpellTextEdit(self)
        #self.set_font_size()
        self.vlayout.addWidget(self.checkbox)
        layout.addLayout(self.vlayout)
        layout.addWidget(self.textfield)
        layout.setContentsMargins(12, 8, 30, 8)
        self.setAutoFillBackground(False)
        self.checkbox.clicked.connect(self.on_checkbox_clicked)

    def set_font_size(self):
        font: QtGui.QFont = self.textfield.font()
        font.setPointSize(font.pointSize() + 2)
        self.textfield.setFont(font)

    @property
    def text(self) -> str:
        return self.textfield.toPlainText()

    def on_checkbox_clicked(self, checked: bool) -> None:
        if checked:
            self.checkbox_clicked.emit(self.textfield.toPlainText())
            self.number_label.setText(str(self._num))
            TextEditItem._num += 1
        else:
            self.number_label.setText('')

    def setText(self, text: str) -> None:
        self.textfield.setPlainText(text)

    def setCheckboxName(self, name: str) -> None:
        self.checkbox.setObjectName(name)

    def sizeHint(self) -> QSize:
        return QSize(400, 180)


class MyTextTemplateEdit(QWidget):
    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)
        self.cif_key: str = ''
        self.ui = text_templates_ui.Ui_TextTemplatesWidget()
        self.ui.setupUi(self)
        self.ui.cancelTextPushButton.clicked.connect(self._on_backbutton_clicked)
        if not self.ui.templatesListWidget.count():
            self.add_more_fiels()
        # self.set_font_size()

    def set_font_size(self):
        textedit_font: QtGui.QFont = self.ui.plainTextEdit.font()
        textedit_font.setPointSize(textedit_font.pointSize() + 3)
        self.ui.plainTextEdit.setFont(textedit_font)
        cif_key_font = self.ui.cifKeyLineEdit.font()
        cif_key_font.setPointSize(textedit_font.pointSize())
        self.ui.cifKeyLineEdit.setFont(cif_key_font)

    def _on_backbutton_clicked(self) -> None:
        self.ui.templatesListWidget.clear()
        TextEditItem._num = 1

    def add_textfields(self, text_list: list | tuple) -> None:
        self.ui.templatesListWidget.clear()
        if text_list:
            for text in text_list:
                self.add_one_textfield(text)
        self.add_more_fiels()

    def add_more_fiels(self) -> None:
        for empty in ('',) * 20:
            self.add_one_textfield(empty)

    def add_one_textfield(self, text: str) -> None:
        edit_item = TextEditItem(self.ui.templatesListWidget)
        edit_item.setText(text)
        edit_item.checkbox_clicked.connect(lambda x: self.ui.plainTextEdit.appendPlainText(x))
        item = QListWidgetItem(self.ui.templatesListWidget)
        item.setSizeHint(edit_item.sizeHint())
        # item.setIcon(qtawesome.icon('fa5.image'))
        self.ui.templatesListWidget.addItem(item)
        self.ui.templatesListWidget.setItemWidget(item, edit_item)

    def get_template_texts(self) -> list[str]:
        texts = []
        for num in range(self.ui.templatesListWidget.count()):
            item = self.ui.templatesListWidget.item(num)
            txt = self.ui.templatesListWidget.itemWidget(item).text
            if txt.strip():
                texts.append(txt)
        return texts

    def clear_fields(self) -> None:
        self.ui.templatesListWidget.clear()
        self.ui.cifKeyLineEdit.clear()
        self.ui.plainTextEdit.clear()


if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication

    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    window = MyTextTemplateEdit()
    # window.add_textfields(txts)
    window.show()
    sys.exit(app.exec())



import re

from qtpy import QtWidgets, QtGui, QtCore
from qtpy.QtCore import Signal

from finalcif.cif import all_cif_dicts
from finalcif.gui.new_key_dialog_ui import Ui_AddKeyWindow


class NewKey(QtWidgets.QMainWindow, Ui_AddKeyWindow):
    new_key_added = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)
        self.searchLineEdit.textChanged.connect(self.search)
        self.addKeyPushButton.clicked.connect(self.add_keys)
        self.cancelPushButton.clicked.connect(lambda: self.close())
        self.keysListWidget.addItems(['', *list(all_cif_dicts.cif_all_dict.keys())])
        self.keysListWidget.item(0).setFlags(
            QtCore.Qt.ItemFlag.ItemIsSelectable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsEditable)
        self.keysListWidget.item(0).setText('Double-click to add custom key.')
        self.keysListWidget.item(0).setForeground(QtGui.QBrush(QtGui.QColor("gray")))
        self.keysListWidget.itemDoubleClicked.connect(self.item_clicked)
        for num in range(self.keysListWidget.count()):
            item = self.keysListWidget.item(num)
            helptext = all_cif_dicts.cif_all_dict.get(item.text())
            item.setToolTip(helptext)

    def item_clicked(self, item: QtWidgets.QListWidgetItem):
        if self.keysListWidget.row(item) == 0:
            self.keysListWidget.item(0).setText('')
            self.keysListWidget.item(0).setForeground(QtGui.QBrush(QtGui.QColor("black")))

    def add_keys(self):
        for item in self.keysListWidget.selectedItems():
            if item.text() and item.text().startswith('_') and len(item.text()) > 2:
                self.new_key_added.emit(item.text())

    def search(self, searchtext: str):
        self.keysListWidget.clear()
        cif_keys = all_cif_dicts.cif_all_dict.keys()
        if searchtext:
            searchpattern = re.compile(f'.*{searchtext}.*', re.IGNORECASE)
            searched = [x for x in cif_keys if searchpattern.match(x)]
            self.keysListWidget.addItems(searched)
        else:
            self.keysListWidget.addItems(cif_keys)


if __name__ == '__main__':
    import sys
    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)

    nk = NewKey()
    nk.show()
    sys.exit(app.exec())



from __future__ import annotations

#   ----------------------------------------------------------------------------
#   "THE BEER-WARE LICENSE" (Revision 42):
#   Daniel Kratzert <dkratzert@gmx.de> wrote this file.  As long as you retain
#   this notice you can do whatever you want with this stuff. If we meet some day,
#   and you think this stuff is worth it, you can buy me a beer in return.
#   ----------------------------------------------------------------------------
import re
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

import gemmi.cif
from qtpy.QtWidgets import QListWidgetItem
from gemmi.cif import Loop, as_string

from finalcif.cif.cif_file_io import CifContainer
from finalcif.cif.text import utf8_to_str, quote, retranslate_delimiter, string_to_utf8
from finalcif.equip_property.tools import read_document_from_cif_file
from finalcif.gui.dialogs import cif_file_save_dialog, show_general_warning, cif_file_open_dialog
from finalcif.gui.finalcif_gui_ui import Ui_FinalCifWindow
from finalcif.tools.misc import grouper
from finalcif.tools.settings import FinalCifSettings

if TYPE_CHECKING:
    from finalcif.appwindow import AppWindow


class AuthorType(Enum):
    publ = 'publ'
    audit = 'audit'

    def __str__(self) -> str:
        return self.name


@dataclass(order=True, init=True)
class Author:
    name: str
    address: str
    email: str
    phone: str
    orcid: str
    footnote: str
    contact_author: bool
    author_type: AuthorType  # 'publ' or 'audit'
    iucr_id: str = ''


class AuthorLoops:
    def __init__(self, ui: Ui_FinalCifWindow, cif: CifContainer, app: AppWindow):
        self.ui = ui
        self.cif = cif
        self.app = app
        self.cif_key = '_publ_author'
        self.keys_list = [f'{self.cif_key}_name', f'{self.cif_key}_address', f'{self.cif_key}_email',
                          f'{self.cif_key}_phone', f'{self.cif_key}_id_orcid', f'{self.cif_key}_id_iucr',
                          f'{self.cif_key}_footnote']
        self.settings = FinalCifSettings()
        if app:
            self.ui.authorEditTabWidget.setCurrentIndex(0)
            self.contact_author_checked(self.ui.ContactAuthorCheckBox.isChecked())
            self.contact_author_checked(self.ui.ContactAuthorCheckBox_cif.isChecked())
            self.connect_signals_and_slots()
            self.show_authors_list()
            self.make_sure_to_save_only_when_name_field_has_text()

    def connect_signals_and_slots(self) -> None:
        self.ui.AddThisAuthorToLoopPushButton.clicked.connect(self.save_author_to_loop)
        self.ui.AddThisAuthorToLoopPushButton_cif.clicked.connect(self.save_author_to_loop)
        self.ui.ContactAuthorCheckBox.stateChanged.connect(self.contact_author_checked)
        self.ui.ContactAuthorCheckBox_cif.stateChanged.connect(self.contact_author_checked)
        self.ui.LoopsTabWidget.currentChanged.connect(
            lambda: self.ui.AddThisAuthorToLoopPushButton.setDisabled(not self.ui.LoopsTabWidget.count()))
        self.ui.authorEditTabWidget.currentChanged.connect(self.set_author_state)
        # self.ui.authorEditTabWidget.currentChanged.connect(self.contact_author_checked)
        self.ui.SaveAuthorLoopToTemplateButton.clicked.connect(self.save_author_to_loop_template)
        self.ui.SaveAuthorLoopToTemplateButton_cif.clicked.connect(self.save_author_to_loop_template)
        # self.ui.authorEditTabWidget.currentChanged.connect(lambda: self.clear_fields())
        self.ui.LoopTemplatesListWidget.clicked.connect(self.load_selected_loop)
        # self.ui.LoopTemplatesListWidget.doubleClicked.connect(self.save_author_to_loop)
        self.ui.DeleteLoopAuthorTemplateButton.clicked.connect(self.delete_current_author)
        self.ui.ExportAuthorPushButton.clicked.connect(self.export_author_template)
        self.ui.ImportAuthorPushButton.clicked.connect(self.import_author)
        self.ui.FullNameLineEdit.textChanged.connect(self.make_sure_to_save_only_when_name_field_has_text)
        self.ui.FullNameLineEdit_cif.textChanged.connect(self.make_sure_to_save_only_when_name_field_has_text)

    def set_author_state(self) -> None:
        if self.ui.authorEditTabWidget.currentWidget().objectName() == 'page_publication':
            self.contact_author_checked(self.ui.ContactAuthorCheckBox.isChecked())
        else:
            self.contact_author_checked(self.ui.ContactAuthorCheckBox_cif.isChecked())

    def make_sure_to_save_only_when_name_field_has_text(self) -> None:
        if not len(self.ui.FullNameLineEdit.text()):
            self.ui.SaveAuthorLoopToTemplateButton.setDisabled(True)
            self.ui.AddThisAuthorToLoopPushButton.setDisabled(True)
        else:
            self.ui.SaveAuthorLoopToTemplateButton.setEnabled(True)
            self.ui.AddThisAuthorToLoopPushButton.setEnabled(True)
        if not len(self.ui.FullNameLineEdit_cif.text()):
            self.ui.SaveAuthorLoopToTemplateButton_cif.setDisabled(True)
            self.ui.AddThisAuthorToLoopPushButton_cif.setDisabled(True)
        else:
            self.ui.SaveAuthorLoopToTemplateButton_cif.setEnabled(True)
            self.ui.AddThisAuthorToLoopPushButton_cif.setEnabled(True)

        if not self.cif:
            self.ui.SaveAuthorLoopToTemplateButton.setDisabled(True)
            self.ui.AddThisAuthorToLoopPushButton.setDisabled(True)

    def get_author_loop(self, author: Author) -> list[str]:
        contact = 'contact_' if author.contact_author else ''
        if author.author_type == AuthorType.publ and author.contact_author:
            author_loop = [f'_{author.author_type}_{contact}author_name',
                           f'_{author.author_type}_{contact}author_address',
                           f'_{author.author_type}_{contact}author_email',
                           f'_{author.author_type}_{contact}author_phone',
                           f'_{author.author_type}_{contact}author_id_orcid',
                           f'_{author.author_type}_{contact}author_id_iucr',
                           ]
        elif author.author_type == AuthorType.publ and not author.contact_author:
            author_loop = [f'_{author.author_type}_{contact}author_name',
                           f'_{author.author_type}_{contact}author_address',
                           f'_{author.author_type}_{contact}author_email',
                           # In fact, email should be disabled in coreCIF version 2.4.5:
                           f'_{author.author_type}_{contact}author_phone',
                           f'_{author.author_type}_{contact}author_id_orcid',
                           f'_{author.author_type}_{contact}author_id_iucr',
                           f'_{author.author_type}_author_footnote']
        elif author.author_type == AuthorType.audit and author.contact_author:
            author_loop = [f'_{author.author_type}_{contact}author_name',
                           f'_{author.author_type}_{contact}author_address',
                           f'_{author.author_type}_{contact}author_email',
                           f'_{author.author_type}_{contact}author_phone']
        else:
            author_loop = [f'_{author.author_type}_{contact}author_name',
                           f'_{author.author_type}_{contact}author_address']
        return author_loop

    def save_author_to_loop(self) -> None:
        author = self.get_author_info()
        row = [author.name, author.address, author.email, author.phone, author.orcid, author.iucr_id, author.footnote]
        author_loop_tags = self.get_author_loop(author)
        table: gemmi.cif.Table = self.cif.block.find(author_loop_tags)
        if table.width():
            gemmi_loop: Loop = table.loop
            if tuple(row) in list(grouper(gemmi_loop.values, gemmi_loop.width())):
                self.app.status_bar.show_message('This author already exists.', 15)
                print('dbg> Author already exists.')
                return
        else:
            gemmi_loop = self.cif.init_loop(author_loop_tags)
        table = self.cif.block.find(author_loop_tags)
        if gemmi_loop.width() < len(row):
            row = row[:gemmi_loop.width()]
        if row not in table:
            table.append_row(row)
        self.app.make_loops_tables()
        self.show_authors_list()

    def get_author_info(self) -> Author:
        if self.ui.authorEditTabWidget.currentWidget().objectName() == 'page_publication':
            name = quote(utf8_to_str(self.ui.FullNameLineEdit.text()))
            address = quote(utf8_to_str(self.ui.AddressTextedit.toPlainText()))
            email = quote(utf8_to_str(self.ui.EMailLineEdit.text()))
            footnote = quote(utf8_to_str(self.ui.FootNoteLineEdit.text()))
            orcid = quote(utf8_to_str(self.ui.ORCIDLineEdit.text()))
            phone = quote(utf8_to_str(self.ui.PhoneLineEdit.text()))
            contact: bool = self.ui.ContactAuthorCheckBox.isChecked()
            iucr = quote(utf8_to_str(self.ui.IUCRIDLineEdit.text()))
            author_type = AuthorType.publ
        else:  # CIF author page:
            name = quote(utf8_to_str(self.ui.FullNameLineEdit_cif.text()))
            address = quote(utf8_to_str(self.ui.AddressTextedit_cif.toPlainText()))
            email = quote(utf8_to_str(self.ui.EMailLineEdit_cif.text()))
            footnote = ''  # No footnote
            orcid = ''  # No ORCID
            phone = quote(utf8_to_str(self.ui.PhoneLineEdit_cif.text()))
            contact: bool = self.ui.ContactAuthorCheckBox_cif.isChecked()
            author_type = AuthorType.audit
            iucr = ''  # No IUCr id
        return Author(name=name, address=address, email=email, footnote=footnote, orcid=orcid,
                      phone=phone, contact_author=contact, author_type=author_type, iucr_id=iucr)

    def set_author_info(self, author: dict[str, str | None] | Author) -> None:
        if not author:
            return
        author_type = AuthorType.publ if self.ui.authorEditTabWidget.currentWidget().objectName() == \
                                         'page_publication' else AuthorType.audit
        if type(author) is dict:
            author = Author(name=author.get('name'), address=author.get('address'), email=author.get('email'),
                            phone=author.get('phone'), orcid=author.get('orcid'), footnote=author.get('footnote'),
                            contact_author=author.get('contact', False), author_type=author_type)
        # In order to use authors for cif and publication:
        author.author_type = author_type
        self.set_authorinfo_from_settings(author)
        self.ui.ContactAuthorCheckBox.setChecked(author.contact_author or False)
        self.ui.ContactAuthorCheckBox_cif.setChecked(author.contact_author or False)

    def set_authorinfo_from_settings(self, author: Author) -> None:
        author_type = "_cif" if author.author_type == AuthorType.audit else ""
        if author.name:
            getattr(self.ui, f'FullNameLineEdit{author_type}').setText(retranslate_delimiter(as_string(author.name)))
        if author.address:
            getattr(self.ui, f'AddressTextedit{author_type}').setText(retranslate_delimiter(as_string(author.address)))
        if author.email:
            getattr(self.ui, f'EMailLineEdit{author_type}').setText(retranslate_delimiter(as_string(author.email)))
        if author.footnote and author.author_type == AuthorType.publ:
            # Audit authors have no footnote:
            self.ui.FootNoteLineEdit.setText(retranslate_delimiter(as_string(author.footnote)))
        if author.orcid and author.author_type == AuthorType.publ:
            # Audit authors have no ORCID:
            self.ui.ORCIDLineEdit.setText(retranslate_delimiter(as_string(author.orcid)))
        if author.iucr_id and author.author_type == AuthorType.publ:
            # Audit authors have no IUCrID:
            self.ui.IUCRIDLineEdit.setText(retranslate_delimiter(as_string(author.iucr_id)))
        if author.phone:
            getattr(self.ui, f'PhoneLineEdit{author_type}').setText(retranslate_delimiter(as_string(author.phone)))

    def clear_fields(self):
        self.ui.FullNameLineEdit.clear()
        self.ui.AddressTextedit.clear()
        self.ui.EMailLineEdit.clear()
        self.ui.FootNoteLineEdit.clear()
        self.ui.ORCIDLineEdit.clear()
        self.ui.PhoneLineEdit.clear()
        self.ui.IUCRIDLineEdit.clear()
        self.ui.ContactAuthorCheckBox.setChecked(False)
        self.ui.FullNameLineEdit_cif.clear()
        self.ui.AddressTextedit_cif.clear()
        self.ui.EMailLineEdit_cif.clear()
        self.ui.PhoneLineEdit_cif.clear()
        self.ui.ContactAuthorCheckBox_cif.setChecked(False)

    def save_author_to_loop_template(self) -> None:
        author = self.get_author_info()
        if not author.name:
            return
        self.general_author_save(author)
        self.clear_fields()

    def general_author_save(self, author: Author) -> None:
        if not author.name:
            return
        if author.contact_author:
            itemtext = f'{string_to_utf8(as_string(author.name))} (contact author)'
        else:
            itemtext = string_to_utf8(as_string(author.name))
        if not itemtext:
            return
        self.settings.save_settings_dict(property='authors_list', name=itemtext, items=author)
        self.show_authors_list()

    def export_author_template(self, filename: str | None = None) -> None:
        """
        Exports the currently selected author to a file.
        """
        selected_template = self.get_currently_selected_author_name()
        if not selected_template:
            return
        blockname = utf8_to_str('__'.join(selected_template.split()))
        template_name = re.sub(r'(?u)[^-\w.]', '_', selected_template)
        if not filename:
            filename = cif_file_save_dialog(f'{template_name}.cif')
        if not filename.strip():
            return
        author_cif = self.put_author_in_cif_object(blockname, filename)
        try:
            author_cif.save(Path(filename))
        except (OSError, PermissionError):
            if Path(filename).is_dir():
                return
            show_general_warning(self, f'No permission to write file to {Path(filename).resolve()}')

    def put_author_in_cif_object(self, blockname: str, filename: str) -> CifContainer:
        author_data = self.author_loopdata(author_name=self.get_selected_loop_name())

        # Convert dict to Author object if needed
        if isinstance(author_data, dict):
            author = Author(
                name=author_data.get('name'),
                address=author_data.get('address'),
                email=author_data.get('email'),
                phone=author_data.get('phone'),
                orcid=author_data.get('orcid'),
                footnote=author_data.get('footnote'),
                contact_author=author_data.get('contact', False),
                author_type=AuthorType.publ,
                iucr_id=author_data.get('iucr_id', '')
            )
        else:
            author = author_data

        data = [author.name, author.address, author.email, author.phone, author.orcid, author.iucr_id,
                author.footnote]
        author_cif = CifContainer(filename, new_block=blockname)
        for key, value in zip(self.keys_list, data, strict=True):
            if value:
                author_cif.set_pair_delimited(key, as_string(value))
        return author_cif

    def import_author(self, filename: str = '') -> None:
        """
        Import an author from a cif file.
        """
        if not filename:
            filename = cif_file_open_dialog(parent=self.app, filter="CIF file (*.cif)")
        if not filename:
            return
        doc = read_document_from_cif_file(filename)
        if not doc:
            return

        for block in doc:
            contact_name = block.find_value('_publ_contact_author_name')
            contact_address = block.find_value('_publ_contact_author_address')

            if not contact_name or contact_name in ('?', '.'):
                contact_author_combined = block.find_value('_publ_contact_author')
                if contact_author_combined and contact_author_combined not in ('?', '.'):
                    ca_str = as_string(contact_author_combined).strip()
                    if ca_str:
                        lines = [ln.strip() for ln in ca_str.splitlines() if ln.strip()]
                        if lines:
                            contact_name = quote(lines[0])
                            if not contact_address and len(lines) > 1:
                                contact_address = quote('\n'.join(lines[1:]))

            if contact_name and contact_name not in ('?', '.'):
                author = Author(
                    name=contact_name,
                    address=contact_address or '',
                    email=block.find_value('_publ_contact_author_email') or '',
                    phone=block.find_value('_publ_contact_author_phone') or '',
                    orcid=block.find_value('_publ_contact_author_id_orcid') or '',
                    iucr_id=block.find_value('_publ_contact_author_id_iucr') or '',
                    footnote=block.find_value('_publ_contact_author_footnote') or '',
                    author_type=AuthorType.publ,
                    contact_author=True
                )
                self.general_author_save(author)

            names = list(block.find_values('_publ_author_name'))
            if names:
                addresses = list(block.find_values('_publ_author_address'))
                emails = list(block.find_values('_publ_author_email'))
                phones = list(block.find_values('_publ_author_phone'))
                orcids = list(block.find_values('_publ_author_id_orcid'))
                iucrs = list(block.find_values('_publ_author_id_iucr'))
                footnotes = list(block.find_values('_publ_author_footnote'))

                for i, name in enumerate(names):
                    if not name or name in ('?', '.'):
                        continue
                    address_raw = addresses[i] if i < len(addresses) and addresses[i] not in ('?', '.') else ''
                    address_stripped = [ln.strip() for ln in address_raw.splitlines() if ln.strip()]
                    author = Author(
                        name=name,
                        address='\n'.join(address_stripped).strip('; '),
                        email=emails[i] if i < len(emails) and emails[i] not in ('?', '.') else '',
                        phone=phones[i] if i < len(phones) and phones[i] not in ('?', '.') else '',
                        orcid=orcids[i] if i < len(orcids) and orcids[i] not in ('?', '.') else '',
                        iucr_id=iucrs[i] if i < len(iucrs) and iucrs[i] not in ('?', '.') else '',
                        footnote=footnotes[i] if i < len(footnotes) and footnotes[i] not in ('?', '.') else '',
                        author_type=AuthorType.publ,
                        contact_author=False
                    )
                    self.general_author_save(author)

        self.show_authors_list()

    def delete_current_author(self) -> None:
        selected_author_name = self.get_currently_selected_author_name()
        if not selected_author_name:
            return
        self.ui.LoopTemplatesListWidget.clear()
        self.settings.delete_template('authors_list', selected_author_name)
        self.show_authors_list()

    def get_selected_loop_name(self) -> str:
        selected_row_text = self.get_currently_selected_author_name() or ''
        return selected_row_text

    def get_currently_selected_author_name(self) -> str:
        return self.ui.LoopTemplatesListWidget.currentIndex().data()

    def load_selected_loop(self) -> None:
        self.clear_fields()
        self.set_author_info(self.author_loopdata(author_name=self.get_selected_loop_name()))
        self.ui.LoopsTabWidget.setCurrentIndex(0)

    def author_loopdata(self, author_name: str) -> dict[str, str] | Author:
        return self.settings.load_settings_dict('authors_list', author_name)

    def show_authors_list(self) -> None:
        self.ui.LoopTemplatesListWidget.clear()
        for author in sorted(self.authors_list()):
            if author:
                self.ui.LoopTemplatesListWidget.addItem(QListWidgetItem(author))

    def export_raw_data(self) -> list[Author]:
        """
        Export all authors in order to export them all"""
        authors = []
        for author in self.authors_list():
            author_data = self.author_loopdata(author)
            authors.append(author_data)
        return authors

    def import_raw_data(self, authors_data: list[Author]) -> None:
        """
        Import all authors from an external file"""
        for author in authors_data:
            if author and isinstance(author, Author):
                self.general_author_save(author)

    def authors_list(self) -> list[str]:
        return self.settings.list_saved_items(property='authors_list')

    def contact_author_checked(self, checked: bool) -> None:
        """
        :parameter checked: state of the ContactAuthorCheckBox
        """
        page = "_cif" if self.ui.authorEditTabWidget.currentWidget().objectName() == 'page_audit' else ""
        author_type = 'publ' if self.ui.authorEditTabWidget.currentWidget().objectName() == 'page_publication' else 'audit'
        contact = 'contact_' if checked else ''

        getattr(self.ui, f'FullNameLineEdit{page}').setToolTip(f'_{author_type}_{contact}author_name')
        getattr(self.ui, f'AddressTextedit{page}').setToolTip(f'_{author_type}_{contact}author_address')
        getattr(self.ui, f'EMailLineEdit{page}').setToolTip(f'_{author_type}_{contact}author_email')
        getattr(self.ui, f'PhoneLineEdit{page}').setToolTip(f'_{author_type}_{contact}author_phone')
        if author_type == AuthorType.publ.name:
            getattr(self.ui, f'ORCIDLineEdit{page}').setToolTip(f'_{author_type}_{contact}author_id_orcid')
            getattr(self.ui, f'IUCRIDLineEdit{page}').setToolTip(f'_{author_type}_{contact}author_id_iucr')
            getattr(self.ui, f'FootNoteLineEdit{page}').setToolTip(f'_{author_type}_{contact}author_footnote')
            getattr(self.ui, f'footnote_label{page}').setToolTip(f'_{author_type}_{contact}author_footnote')
            if checked:
                getattr(self.ui, f'FootNoteLineEdit{page}').setDisabled(True)
                getattr(self.ui, f'footnote_label{page}').setDisabled(True)
                # In fact, this should be enabled in coreCIF version 2.4.5:
                # getattr(self.ui, f'PhoneLineEdit{page}').setEnabled(True)
                # getattr(self.ui, f'PhoneLabel{page}').setEnabled(True)
            else:
                # getattr(self.ui, f'PhoneLineEdit{page}').setDisabled(True)
                # getattr(self.ui, f'PhoneLabel{page}').setDisabled(True)
                getattr(self.ui, f'FootNoteLineEdit{page}').setEnabled(True)
                getattr(self.ui, f'footnote_label{page}').setEnabled(True)
        if not checked and author_type == AuthorType.audit.name:
            getattr(self.ui, f'PhoneLineEdit{page}').setDisabled(True)
            getattr(self.ui, f'EMailLineEdit{page}').setDisabled(True)
            getattr(self.ui, f'EmailLabel{page}').setDisabled(True)
            getattr(self.ui, f'PhoneLabel{page}').setDisabled(True)
        elif author_type == AuthorType.audit.name:
            getattr(self.ui, f'PhoneLineEdit{page}').setEnabled(True)
            getattr(self.ui, f'EMailLineEdit{page}').setEnabled(True)
            getattr(self.ui, f'EmailLabel{page}').setEnabled(True)
            getattr(self.ui, f'PhoneLabel{page}').setEnabled(True)


if __name__ == '__main__':
    import pprint

    l = AuthorLoops(Ui_FinalCifWindow(), CifContainer('test-data/1000007.cif'), None)
    data = l.export_raw_data()
    pprint.pprint(data)




#  ----------------------------------------------------------------------------
#  "THE BEER-WARE LICENSE" (Revision 42):
#  dkratzert@gmx.de> wrote this file.  As long as you retain
#  this notice you can do whatever you want with this stuff. If we meet some day,
#  and you think this stuff is worth it, you can buy me a beer in return.
#  Dr. Daniel Kratzert
#  ----------------------------------------------------------------------------
import os

os.environ['QT_API'] = 'pyside6'
import sys
import time
import traceback
from pathlib import Path
from qtpy.QtGui import QIcon
from finalcif import VERSION
from finalcif.app_path import application_path
from finalcif.appwindow import AppWindow, DEBUG, app
from finalcif.gui.dialogs import show_bug_found_warning, show_general_warning


def my_exception_hook(exctype: type[BaseException], value: BaseException, error_traceback: traceback,
                      exit=True) -> None:
    """
    Hooks into Exceptions to create debug reports.
    """
    errortext = (f'FinalCif V{VERSION} crash report\n\n'
                 f'Please send also the corresponding CIF file, if possible. \n'
                 f'Python {sys.version}\n'
                 f'Platform: {sys.platform}\n'
                 f'Date: {time.asctime(time.localtime(time.time()))}\n'
                 f'Finalcif crashed during the following operation:\n\n'
                 f'{"-" * 120}\n'
                 # f'{"".join(traceback.format_tb(error_traceback))}\n'
                 # f'{str(exctype.__name__)}: '
                 # f'{str(value)} \n'
                 # f'{"-" * 120}\n'
                 )

    # Walk through the traceback and extract local variables
    for frame, _ in traceback.walk_tb(error_traceback):
        errortext += (f'File "{frame.f_code.co_filename}", line {frame.f_lineno}, in '
                      f'{frame.f_code.co_qualname}(...):\n')
        newline = '\n'
        errortext += '  Locals: \n    '
        parts = []
        for k, v in frame.f_locals.items():
            try:
                v_repr = repr(v)
            except Exception:
                v_repr = '<unrepresentable>'
            parts.append(
                f"  {k}:{newline}          "
                f"{'          '.join([x + newline for x in v_repr.splitlines()])}"
            )
        errortext += "    ".join(parts) + '\n\n'

    errortext += f'{"-" * 120}\n'
    errortext += f'{exctype.__name__!s}: {value!s} \n'
    errortext += f'{"-" * 120}\n'

    logfile = Path.home().joinpath(Path(r'finalcif-crash.txt'))
    try:
        logfile.write_text(errortext, encoding='utf-8', errors='ignore')
    except PermissionError:
        pass
    sys.__excepthook__(exctype, value, error_traceback)
    show_bug_found_warning(logfile)
    if exit:
        sys.exit(1)


def main():
    if not DEBUG:
        sys.excepthook = my_exception_hook
    # windows_style = QStyleFactory.create('Fusion')
    # app.setStyle(windows_style)
    file = None
    if len(sys.argv) > 1:
        file = Path(sys.argv[1])
        if not file.is_file():
            show_general_warning(parent=None,
                                 warn_text=f'The file {file.resolve().absolute()} \nyou tried to open does not exist.',
                                 window_title='File not found')
            file = None
    app.setQuitOnLastWindowClosed(True)
    w = AppWindow(file=file)
    app.setWindowIcon(QIcon(str(application_path / r'icon/finalcif2.png')))
    w.setWindowTitle(f'FinalCif v{VERSION}')
    # w.showMaximized()  # For full screen view
    w.setBaseSize(1200, 780)
    sys.exit(app.exec())


if __name__ == '__main__':
    main()



import shutil
import unittest
from pathlib import Path

from finalcif.cif.cif_file_io import CifContainer

data = Path('.')


class CifFileCRCTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifContainer(data / 'tests/examples/1979688.cif')

    def test_calc_crc(self):
        self.assertEqual(20714, self.cif.calc_checksum(self.cif['_shelx_hkl_file']))


class CifFileCRClargerTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifContainer(data / 'test-data/DK_Zucker2_0m.cif')

    def test_calc_crc(self):
        self.assertEqual(26780, self.cif.calc_checksum(self.cif['_shelx_hkl_file']))


class CifFileTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifContainer(data / 'tests/examples/1979688.cif')

    def test_calc_crc(self):
        self.assertEqual(3583, self.cif.calc_checksum('hello world'))

    def test_res_crc(self):
        self.assertEqual(17612, self.cif.res_checksum_calcd)

    def test_hkl_crc(self):
        self.assertEqual(20714, self.cif.hkl_checksum_calcd)

    def test_res_crc_without_res(self):
        self.assertEqual(0, CifContainer(data / 'test-data/1000006.cif').res_checksum_calcd)

    def test_get_unknown_value_from_key(self):
        self.assertEqual('', self.cif['_chemical_melting_point'])

    def test_get_known_value_from_key(self):
        self.assertEqual('702.70', self.cif['_chemical_formula_weight'])

    def test_get_spgr(self):
        self.assertEqual('P 21 21 2', self.cif.space_group)

    def test_symmops(self):
        self.assertEqual(['x, y, z', '-x, -y, z', '-x+1/2, y+1/2, -z', 'x+1/2, -y+1/2, -z'], self.cif.symmops)

    def test_symmops_from_spgr(self):
        self.assertEqual(['x,y,z', '-x,-y,z', 'x+1/2,-y+1/2,-z', '-x+1/2,y+1/2,-z'], self.cif.symmops_from_spgr)

    def test_centrosymm(self):
        self.assertEqual(False, self.cif.is_centrosymm)
        c = CifContainer(data / 'test-data/DK_ML7-66-final.cif')
        self.assertEqual(True, c.is_centrosymm)

    def test_ishydrogen(self):
        self.assertEqual(True, self.cif.ishydrogen('H18a'))
        self.assertEqual(True, self.cif.ishydrogen('H18A'))
        self.assertEqual(False, self.cif.ishydrogen('C2'))
        self.assertEqual(False, self.cif.ishydrogen('c2'))

    def test_cell(self):
        expected = [round(x, 8) for x in (19.678, 37.02290000000001, 4.772, 90.0, 90.0, 90.0, 3476.576780226401)]
        actual = [round(y, 8) for y in self.cif.cell]
        self.assertEqual(expected, actual)

    def test_natoms(self):
        self.assertEqual(94, self.cif.natoms())
        self.assertEqual(52, self.cif.natoms(without_h=True))

    def test_checksum_tests(self):
        self.assertEqual(True, self.cif.test_hkl_checksum())
        self.assertEqual(True, self.cif.test_res_checksum())

    def test_checksum_test_without_checksum(self):
        self.assertEqual(True, CifContainer(data / 'test-data/1000006.cif').test_res_checksum())
        self.assertEqual(True, CifContainer(data / 'test-data/1000006.cif').test_hkl_checksum())

    def test_distance_from_string(self):
        self.assertEqual('1.527(3)', self.cif.bond_dist('C1-C2'))

    def test_distance_from_false_string(self):
        self.assertEqual(None, self.cif.bond_dist('C1-C999'))

    def test_angle_from_string(self):
        self.assertEqual('109.9', self.cif.angle('C1-C2-H2'))

    def test_angle_from_false_string(self):
        self.assertEqual(None, self.cif.angle('C1-C2-Hxx'))

    def test_torsion_angle_from_string(self):
        self.assertEqual('173.5(2)', self.cif.torsion('C1-C2-C3-C4'))

    def test_torsion_angle_from_false_string(self):
        self.assertEqual(None, self.cif.torsion('C1-C2-C3-C999'))

    def test_add_loop(self):
        self.assertEqual(9, self.cif.n_loops)
        self.cif.add_loop_to_cif(loop_tags=['_foo', '_bar'], row_values=['fooval', 'barval'])
        self.assertEqual(10, self.cif.n_loops)
        self.assertEqual(['fooval', 'barval'], self.cif.get_loop('_foo').values)

    def test_add_complete_loop(self):
        loop_data = {'_foo_bar': ['1', '2', '3', '4'], '_baz_hhk': ['omega', 'phi', 'phi', 'chi']}
        self.cif.add_loop_from_columns(list(loop_data.keys()), list(loop_data.values()))
        self.assertEqual(['1', 'omega', '2', 'phi', '3', 'phi', '4', 'chi'],
                         self.cif.get_loop('_foo_bar').values)

    def test_hklf(self):
        self.assertEqual(4, self.cif.hklf_number)

    def test_hkl_as_cif(self):
        result = ('data_cu_BruecknerJK_153F40_0m\n'
                  'loop_\n'
                  '_refln_index_h\n'
                  '_refln_index_k\n'
                  '_refln_index_l\n'
                  '_refln_F_squared_meas\n'
                  '_refln_F_squared_sigma\n'
                  '_refln_scale_group_code\n'
                  '1 0 0 0.36031 0.34981 12\n'
                  '-1 0 0 -0.0279 0.03389 7\n')
        self.assertEqual(result, self.cif.hkl_as_cif[:200])

    def test_publ_flag_not_set(self):
        self.assertEqual("bond(label1='C1', label2='O1', dist='1.438(3)', symm='.')", str(next(iter(self.cif.bonds()))))
        self.cif.block.find_loop('_geom_bond_publ_flag')[0] = 'no'
        # This is not C1-O1, but C1-C14, because the first bond is with publ_flag 'no'.
        self.assertEqual("bond(label1='C1', label2='C14', dist='1.519(3)', symm='.')", str(next(iter(self.cif.bonds()))))
        self.cif.block.find_loop('_geom_bond_publ_flag').erase()
        self.assertEqual([], list(self.cif.block.find_loop('_geom_bond_publ_flag')))
        self.assertEqual("bond(label1='C1', label2='O1', dist='1.438(3)', symm='.')", str(next(iter(self.cif.bonds()))))


class TestQuotationMark(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifContainer(data / 'tests/examples/1979688.cif')
        shutil.copyfile(self.cif.fileobj, data / 'test.cif')
        self.cif2 = CifContainer(data / 'test.cif')

    def tearDown(self) -> None:
        self.cif2.fileobj.unlink(missing_ok=True)

    def test_insert_quotation_mark(self):
        self.cif['_diffrn_detector'] = "Jesus' live"
        self.assertEqual("Jesus' live", self.cif['_diffrn_detector'])
        self.cif.save(self.cif2.fileobj)
        self.cif2 = CifContainer(self.cif2.fileobj)
        self.assertEqual("Jesus' live", self.cif['_diffrn_detector'])


class TestMultiCif(unittest.TestCase):
    def setUp(self) -> None:
        self.cif = CifContainer(data / 'tests/examples/multi.cif')

    def test_ismulti(self):
        self.assertEqual(True, self.cif.is_multi_cif)

    def test_isempty(self):
        self.assertEqual(False, self.cif.is_empty())


class TestVrfKeyOrdering(unittest.TestCase):
    """Tests that _vrf_* keys are moved to the top by order_cif_keys() and keys_with_essentials()."""

    def setUp(self) -> None:
        self.cif = CifContainer(data / 'tests/examples/work/cu_BruecknerJK_153F40_0m.cif')

    def test_order_cif_keys_puts_vrf_first(self):
        """After order_cif_keys(), the first key in the block must be a _vrf_* key."""
        self.cif.order_cif_keys()
        keys = self.cif.keys()
        vrf_keys = [k for k in keys if k.startswith('_vrf')]
        self.assertTrue(len(vrf_keys) > 0, 'Test CIF should contain at least one _vrf_* key')
        # All _vrf_* keys should appear before any non-_vrf_* key
        first_non_vrf_index = next((i for i, k in enumerate(keys) if not k.startswith('_vrf')), len(keys))
        last_vrf_index = max((i for i, k in enumerate(keys) if k.startswith('_vrf')), default=-1)
        self.assertLess(last_vrf_index, first_non_vrf_index,
                        '_vrf_* keys should all appear before non-_vrf_* keys after ordering')

    def test_keys_with_essentials_puts_vrf_first_in_questions(self):
        """_vrf_* question keys must appear before non-_vrf_* question keys in the table."""
        questions, _ = self.cif.keys_with_essentials()
        vrf_indices = [i for i, (k, _v) in enumerate(questions) if k.startswith('_vrf')]
        non_vrf_indices = [i for i, (k, _v) in enumerate(questions) if not k.startswith('_vrf')]
        if vrf_indices and non_vrf_indices:
            self.assertLess(max(vrf_indices), min(non_vrf_indices),
                            '_vrf_* question keys should precede non-_vrf_* question keys')

    def test_keys_with_essentials_puts_vrf_first_in_with_values(self):
        """_vrf_* answered keys must appear before non-_vrf_* answered keys in the table."""
        _, with_values = self.cif.keys_with_essentials()
        vrf_indices = [i for i, (k, _v) in enumerate(with_values) if k.startswith('_vrf')]
        non_vrf_indices = [i for i, (k, _v) in enumerate(with_values) if not k.startswith('_vrf')]
        if vrf_indices and non_vrf_indices:
            self.assertLess(max(vrf_indices), min(non_vrf_indices),
                            '_vrf_* answered keys should precede non-_vrf_* answered keys')


if __name__ == '__main__':
    unittest.main()



import json
import os
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import MagicMock, patch

from finalcif.tools.settings import (
    FinalCifSettings,
    _convert_qsettings_value,
    _migrate_qsettings_to_json,
    load_template_file,
)


class TestFinalCifSettings(TestCase):

    def setUp(self) -> None:
        os.environ["RUNNING_TEST"] = 'True'
        # Create a temp JSON file with test fixture data
        self.tmpdir = tempfile.mkdtemp()
        self.settings_path = Path(self.tmpdir) / 'test_settings.json'
        fixture = {
            'property': {
                '_chemical_absolute_configuration': [
                    '_chemical_absolute_configuration',
                    ['', 'rm', 'ad', 'rmad', 'syn', 'unk']
                ],
                '_exptl_absorpt_correction_type': [
                    '_exptl_absorpt_correction_type',
                    ['', 'analytical', 'cylinder', 'empirical', 'gaussian', 'integration',
                     'multi-scan', 'none', 'numerical', 'psi-scan', 'refdelf', 'sphere']
                ],
                '_diffrn_ambient_environment': [
                    '_diffrn_ambient_environment',
                    ['', 'N~2~', 'He', 'vacuum', 'mother liquor', 'Ar', 'H~2~']
                ],
            }
        }
        self.settings_path.write_text(json.dumps(fixture), encoding='utf-8')
        self.s = FinalCifSettings(settings_path=self.settings_path)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_load_property_values_by_key(self):
        result = [(0, ''), (1, 'N~2~'), (2, 'He'), (3, 'vacuum'), (4, 'mother liquor'), (5, 'Ar'), (6, 'H~2~')]
        self.assertEqual(result, self.s.load_property_values_by_key('_diffrn_ambient_environment'))

    def test_load_cif_keys_of_properties(self):
        self.assertEqual(['_chemical_absolute_configuration', '_exptl_absorpt_correction_type'],
                         self.s.load_cif_keys_of_properties()[:2])

    def test_load_property_values_by_key_empty(self):
        self.assertEqual([(0, '')], self.s.load_property_values_by_key(''))


class TestSettingsSaveLoad(TestCase):

    def setUp(self) -> None:
        self.tmpdir = tempfile.mkdtemp()
        self.settings_path = Path(self.tmpdir) / 'test_settings.json'
        self.s = FinalCifSettings(settings_path=self.settings_path)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def test_save_and_load_key_value(self):
        self.s.save_key_value('test_key', 'test_value')
        self.assertEqual('test_value', self.s.load_value_of_key('test_key'))

    def test_save_and_load_options(self):
        opts = {'report_text': False, 'picture_width': 5.0}
        self.s.save_options(opts)
        loaded = self.s.load_options()
        self.assertFalse(loaded['report_text'])
        self.assertEqual(5.0, loaded['picture_width'])

    def test_save_and_load_settings_list(self):
        self.s.save_settings_list('equipment', 'TestEquip', [['key1', 'val1'], ['key2', 'val2']])
        result = self.s.load_settings_list('equipment', 'TestEquip')
        self.assertEqual([['key1', 'val1'], ['key2', 'val2']], result)

    def test_delete_template(self):
        self.s.save_settings_list('equipment', 'ToDelete', [['a', 'b']])
        self.assertIn('ToDelete', self.s.get_equipment_list())
        self.s.delete_template('equipment', 'ToDelete')
        self.assertNotIn('ToDelete', self.s.get_equipment_list())

    def test_save_and_load_workdir(self):
        self.s.save_current_dir('/some/path')
        self.assertEqual('/some/path', self.s.load_last_workdir())

    def test_recent_files(self):
        self.s.save_recent_files(['/a.cif', '/b.cif'])
        self.assertEqual(['/a.cif', '/b.cif'], self.s.load_recent_files())

    def test_empty_recent_files(self):
        self.s.save_recent_files([])
        self.assertEqual([], self.s.load_recent_files())

    def test_property_list(self):
        self.assertIsNone(self.s.load_property_list())
        self.s.save_key_value('property_list', ['a', 'b'])
        self.assertEqual(['a', 'b'], self.s.load_property_list())

    def test_json_persistence(self):
        """Verify settings survive a new FinalCifSettings instance."""
        self.s.save_key_value('persist_test', 42)
        self.s.flush()
        s2 = FinalCifSettings(settings_path=self.settings_path)
        self.assertEqual(42, s2.load_value_of_key('persist_test'))

    def test_load_settings_dict(self):
        self.s.save_settings_dict('authors_list', 'Jane', {'name': 'Jane', 'email': 'j@e.com'})
        result = self.s.load_settings_dict('authors_list', 'Jane')
        self.assertEqual({'name': 'Jane', 'email': 'j@e.com'}, result)

    def test_load_settings_dict_returns_non_dict_values(self):
        """Non-dict stored values (e.g. Author dataclass) must not be discarded."""
        # Simulate storing a non-dict value (e.g. a list) – load_settings_dict
        # must return it unchanged rather than returning an empty dict.
        self.s.save_settings_dict('misc', 'key', ['a', 'b', 'c'])
        result = self.s.load_settings_dict('misc', 'key')
        self.assertEqual(['a', 'b', 'c'], result)

    def test_load_settings_dict_returns_empty_dict_when_missing(self):
        """Missing key should still return {}."""
        result = self.s.load_settings_dict('authors_list', 'NonExistent')
        self.assertEqual({}, result)

    def test_default_options(self):
        """When no options are saved, defaults should be returned."""
        opts = self.s.load_options()
        self.assertTrue(opts['report_text'])
        self.assertEqual(7.5, opts['picture_width'])
        self.assertTrue(opts['atoms_table'])

    def test_saved_options_override_defaults(self):
        """Saved option values should be returned instead of defaults."""
        self.s.save_options({'report_text': False, 'picture_width': 3.0, 'without_h': True})
        opts = self.s.load_options()
        self.assertFalse(opts['report_text'])
        self.assertEqual(3.0, opts['picture_width'])
        self.assertTrue(opts['without_h'])
        # Hard-coded overrides should still be present
        self.assertTrue(opts['atoms_table'])


# ---------------------------------------------------------------------------
# Helpers shared by migration tests
# ---------------------------------------------------------------------------

def _make_mock_qsettings(top_level_keys: dict, groups: dict) -> MagicMock:
    """Build a MagicMock that behaves like QSettings.

    Parameters
    ----------
    top_level_keys : dict
        Keys (without a group prefix) and their values.
    groups : dict
        Mapping of group name → {key: value}.  beginGroup / endGroup keep
        track of which group is current.
    """
    qs = MagicMock()
    current_group: list = [None]

    def begin_group(g: str) -> None:
        current_group[0] = g

    def end_group() -> None:
        current_group[0] = None

    def all_keys() -> list:
        if current_group[0] is not None:
            return list(groups.get(current_group[0], {}).keys())
        return [k for k in top_level_keys if '/' not in k]

    def child_groups() -> list:
        return list(groups.keys())

    def value(key: str):
        if current_group[0] is not None:
            return groups[current_group[0]][key]
        return top_level_keys[key]

    qs.beginGroup.side_effect = begin_group
    qs.endGroup.side_effect = end_group
    qs.allKeys.side_effect = all_keys
    qs.childGroups.side_effect = child_groups
    qs.value.side_effect = value
    return qs


# ---------------------------------------------------------------------------
# Tests for _convert_qsettings_value
# ---------------------------------------------------------------------------

class TestConvertQSettingsValue(TestCase):
    """Unit tests for _convert_qsettings_value.

    No file I/O or AppWindow required.
    """

    def test_none_returns_none(self):
        self.assertIsNone(_convert_qsettings_value(None))

    def test_string_returned_unchanged(self):
        self.assertEqual('hello world', _convert_qsettings_value('hello world'))

    def test_integer_returned_unchanged(self):
        self.assertEqual(42, _convert_qsettings_value(42))

    def test_float_returned_unchanged(self):
        self.assertAlmostEqual(3.14, _convert_qsettings_value(3.14))

    def test_bool_returned_unchanged(self):
        self.assertIs(True, _convert_qsettings_value(True))
        self.assertIs(False, _convert_qsettings_value(False))

    def test_list_of_strings_returned_as_list(self):
        result = _convert_qsettings_value(['a', 'b', 'c'])
        self.assertEqual(['a', 'b', 'c'], result)

    def test_tuple_converted_to_list(self):
        result = _convert_qsettings_value(('x', 'y'))
        self.assertEqual(['x', 'y'], result)

    def test_nested_list_recurses(self):
        result = _convert_qsettings_value([1, 'two', [3]])
        self.assertEqual([1, 'two', [3]], result)

    def test_dict_converted_recursively(self):
        result = _convert_qsettings_value({'key': 'val', 'num': 7})
        self.assertEqual({'key': 'val', 'num': 7}, result)

    def test_unknown_type_returns_string(self):
        class _Weird:
            def __str__(self):
                return 'weird_repr'
        result = _convert_qsettings_value(_Weird())
        self.assertEqual('weird_repr', result)

    def test_qpoint_serialised_with_type_tag(self):
        from qtpy.QtCore import QPoint
        result = _convert_qsettings_value(QPoint(10, 20))
        self.assertEqual({'__type__': 'QPoint', 'x': 10, 'y': 20}, result)

    def test_qsize_serialised_with_type_tag(self):
        from qtpy.QtCore import QSize
        result = _convert_qsettings_value(QSize(800, 600))
        self.assertEqual({'__type__': 'QSize', 'width': 800, 'height': 600}, result)

    def test_list_of_text_templates_unchanged(self):
        """A list of template strings (the most common QSettings value for text_templates) survives."""
        templates = [
            'All H-atoms were refined freely.',
            'H atoms were placed in calculated positions.',
            'Multi-line\nvalue with newline.',
        ]
        result = _convert_qsettings_value(templates)
        self.assertEqual(templates, result)


# ---------------------------------------------------------------------------
# Tests for _migrate_qsettings_to_json
# ---------------------------------------------------------------------------

class TestMigrateQSettingsToJson(TestCase):
    """Tests for the one-time QSettings → JSON migration helper.

    QSettings is mocked so the test never touches the real application store.
    """

    def setUp(self) -> None:
        self.tmpdir = Path(tempfile.mkdtemp())
        self.json_path = self.tmpdir / 'settings.json'

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def _run_migration(self, top_level_keys: dict, groups: dict) -> dict:
        mock_qs = _make_mock_qsettings(top_level_keys, groups)
        with patch('qtpy.QtCore.QSettings', return_value=mock_qs):
            return _migrate_qsettings_to_json(self.json_path)

    def test_migrated_flag_is_set(self):
        result = self._run_migration({}, {})
        self.assertTrue(result.get('_migrated_from_qsettings'))

    def test_top_level_key_migrated(self):
        result = self._run_migration({'recent_files': ['/a.cif', '/b.cif']}, {})
        self.assertEqual(['/a.cif', '/b.cif'], result.get('recent_files'))

    def test_text_templates_group_migrated(self):
        """text_templates stored in QSettings must appear in the migration output."""
        groups = {
            'text_templates': {
                '_refine_special_details': [
                    'All H-atom positions were refined freely.',
                    'H atoms constrained to C-H 0.96 Å.',
                ],
                '_exptl_crystal_description': ['Colourless prism'],
            }
        }
        result = self._run_migration({}, groups)
        tt = result.get('text_templates', {})
        self.assertIn('_refine_special_details', tt)
        self.assertEqual(2, len(tt['_refine_special_details']))
        self.assertEqual('All H-atom positions were refined freely.', tt['_refine_special_details'][0])
        self.assertIn('_exptl_crystal_description', tt)

    def test_multiple_groups_all_migrated(self):
        """All QSettings groups are migrated, not just text_templates."""
        groups = {
            'text_templates': {'_key_a': ['text a']},
            'equipment': {'MyDiffractometer': [['_diffrn_source', 'Mo']]},
            'MainWindow': {'maximized': 'false'},
        }
        result = self._run_migration({}, groups)
        self.assertIn('text_templates', result)
        self.assertIn('equipment', result)
        self.assertIn('MainWindow', result)

    def test_empty_qsettings_produces_only_migrated_flag(self):
        result = self._run_migration({}, {})
        self.assertEqual({'_migrated_from_qsettings': True}, result)

    def test_unicode_text_templates_preserved(self):
        unicode_templates = ['Δρ_max = 0.15 e Å⁻³', 'α-phase', 'μ = 0.2 mm⁻¹']
        groups = {'text_templates': {'_refine_ls_extinction_coef': unicode_templates}}
        result = self._run_migration({}, groups)
        migrated = result['text_templates']['_refine_ls_extinction_coef']
        self.assertEqual(unicode_templates, migrated)

    def test_multiline_strings_in_templates_preserved(self):
        multiline = ['First line\nSecond line\nThird line', 'Another entry']
        groups = {'text_templates': {'_refine_special_details': multiline}}
        result = self._run_migration({}, groups)
        self.assertEqual(multiline, result['text_templates']['_refine_special_details'])

    def test_qpoint_value_migrated_with_type_tag(self):
        from qtpy.QtCore import QPoint
        result = self._run_migration({'win_pos': QPoint(50, 100)}, {})
        self.assertEqual({'__type__': 'QPoint', 'x': 50, 'y': 100}, result.get('win_pos'))

    def test_qsize_value_migrated_with_type_tag(self):
        from qtpy.QtCore import QSize
        groups = {'MainWindow': {'size': QSize(1024, 768)}}
        result = self._run_migration({}, groups)
        self.assertEqual({'__type__': 'QSize', 'width': 1024, 'height': 768},
                         result['MainWindow']['size'])


# ---------------------------------------------------------------------------
# Tests for FinalCifSettings auto-migration on first launch
# ---------------------------------------------------------------------------

class TestFinalCifSettingsAutoMigration(TestCase):
    """Tests that FinalCifSettings triggers the QSettings migration exactly
    when (and only when) no JSON file exists yet.
    """

    def setUp(self) -> None:
        self.tmpdir = Path(tempfile.mkdtemp())
        self.settings_path = self.tmpdir / 'settings.json'

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    def _make_settings_with_mock_qs(self, top_level_keys: dict, groups: dict) -> FinalCifSettings:
        mock_qs = _make_mock_qsettings(top_level_keys, groups)
        with patch('qtpy.QtCore.QSettings', return_value=mock_qs):
            return FinalCifSettings(settings_path=self.settings_path)

    def test_migration_runs_when_json_is_absent(self):
        """JSON must be created from QSettings when no JSON file exists."""
        groups = {'text_templates': {'_key': ['template text']}}
        s = self._make_settings_with_mock_qs({}, groups)
        self.assertTrue(self.settings_path.exists())

    def test_migration_json_contains_migrated_flag(self):
        s = self._make_settings_with_mock_qs({}, {})
        raw = json.loads(self.settings_path.read_text())
        self.assertTrue(raw.get('_migrated_from_qsettings'))

    def test_migration_text_templates_accessible_via_settings_api(self):
        """After migration, text templates can be read through the normal API."""
        templates = ['Refined freely.', 'H constrained to C-H 0.96.']
        groups = {'text_templates': {'_refine_special_details': templates}}
        s = self._make_settings_with_mock_qs({}, groups)

        loaded = s.load_settings_list('text_templates', '_refine_special_details')
        self.assertEqual(templates, loaded)

    def test_migration_multiple_text_template_keys_all_accessible(self):
        groups = {
            'text_templates': {
                '_refine_special_details': ['text A', 'text B'],
                '_exptl_crystal_description': ['Red block', 'Colourless prism'],
                '_vrf_DIFF003_somecif': ['Response to alert DIFF003.'],
            }
        }
        s = self._make_settings_with_mock_qs({}, groups)
        self.assertEqual(2, len(s.load_settings_list('text_templates', '_refine_special_details')))
        self.assertEqual(['Red block', 'Colourless prism'],
                         s.load_settings_list('text_templates', '_exptl_crystal_description'))
        self.assertIn('_vrf_DIFF003_somecif', s.list_saved_items('text_templates'))

    def test_migration_skipped_when_json_already_exists(self):
        """If the JSON already exists, QSettings must never be instantiated."""
        self.settings_path.write_text(
            json.dumps({'text_templates': {'_key': ['pre-existing data']}}),
            encoding='utf-8',
        )
        mock_qs_class = MagicMock()
        with patch('qtpy.QtCore.QSettings', mock_qs_class):
            s = FinalCifSettings(settings_path=self.settings_path)

        mock_qs_class.assert_not_called()
        self.assertEqual(['pre-existing data'], s.load_settings_list('text_templates', '_key'))

    def test_migration_json_persists_across_new_instance(self):
        """Migrated data survives a second FinalCifSettings instantiation."""
        groups = {'text_templates': {'_key': ['migrated data']}}
        self._make_settings_with_mock_qs({}, groups)

        # Re-open without mocking – should read the JSON created by migration
        s2 = FinalCifSettings(settings_path=self.settings_path)
        self.assertEqual(['migrated data'], s2.load_settings_list('text_templates', '_key'))


# ---------------------------------------------------------------------------
# Tests for text template CRUD via FinalCifSettings (no migration involved)
# ---------------------------------------------------------------------------

class TestTextTemplateSettings(TestCase):
    """Tests for saving, loading, listing, and deleting text templates.

    Uses an isolated temp JSON file – no AppWindow or QSettings required.
    """

    def setUp(self) -> None:
        self.tmpdir = Path(tempfile.mkdtemp())
        self.settings_path = self.tmpdir / 'settings.json'
        self.s = FinalCifSettings(settings_path=self.settings_path)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    # --- save / load ---

    def test_save_and_load_text_template(self):
        items = ['Template line A.', 'Template line B.']
        self.s.save_settings_list('text_templates', '_refine_special_details', items)
        self.assertEqual(items, self.s.load_settings_list('text_templates', '_refine_special_details'))

    def test_multiline_string_in_template_preserved(self):
        items = ['First line\nSecond line\nThird line', 'Another entry']
        self.s.save_settings_list('text_templates', '_exptl_crystal_description', items)
        self.assertEqual(items, self.s.load_settings_list('text_templates', '_exptl_crystal_description'))

    def test_unicode_characters_in_template_preserved(self):
        items = [
            'Δρ_max = 0.15 e Å⁻³',  # Greek and superscript
            'α-phase transition',
            'μ = 0.12 mm⁻¹',
        ]
        self.s.save_settings_list('text_templates', '_refine_ls_extinction_coef', items)
        self.assertEqual(items, self.s.load_settings_list('text_templates', '_refine_ls_extinction_coef'))

    def test_cif_delimiter_chars_in_template_preserved(self):
        """CIF tilde/backslash delimiter sequences are plain strings in the template and must survive JSON."""
        items = ['N~2~ atmosphere', r'\a special char', 'normal text']
        self.s.save_settings_list('text_templates', '_diffrn_ambient_environment', items)
        self.assertEqual(items, self.s.load_settings_list('text_templates', '_diffrn_ambient_environment'))

    def test_empty_template_list_saved_and_loaded(self):
        self.s.save_settings_list('text_templates', '_key_empty', [])
        self.assertEqual([], self.s.load_settings_list('text_templates', '_key_empty'))

    def test_missing_key_returns_empty_list(self):
        result = self.s.load_settings_list('text_templates', '_nonexistent_key')
        self.assertEqual([], result)

    def test_overwrite_existing_template(self):
        self.s.save_settings_list('text_templates', '_key_a', ['old'])
        self.s.save_settings_list('text_templates', '_key_a', ['new1', 'new2'])
        self.assertEqual(['new1', 'new2'], self.s.load_settings_list('text_templates', '_key_a'))

    # --- list / delete ---

    def test_list_saved_text_templates(self):
        self.s.save_settings_list('text_templates', '_key_1', ['a'])
        self.s.save_settings_list('text_templates', '_key_2', ['b'])
        items = self.s.list_saved_items('text_templates')
        self.assertIn('_key_1', items)
        self.assertIn('_key_2', items)

    def test_delete_text_template_removes_key(self):
        self.s.save_settings_list('text_templates', '_to_delete', ['data'])
        self.assertIn('_to_delete', self.s.list_saved_items('text_templates'))
        self.s.delete_template('text_templates', '_to_delete')
        self.assertNotIn('_to_delete', self.s.list_saved_items('text_templates'))

    def test_delete_nonexistent_template_is_silent(self):
        """Deleting a key that does not exist must not raise."""
        self.s.delete_template('text_templates', '_nonexistent')

    # --- persistence ---

    def test_template_persists_across_new_settings_instance(self):
        self.s.save_settings_list('text_templates', '_key_persist', ['persistent data'])
        self.s.flush()
        s2 = FinalCifSettings(settings_path=self.settings_path)
        self.assertEqual(['persistent data'], s2.load_settings_list('text_templates', '_key_persist'))

    # --- VRF keys ---

    def test_vrf_key_stored_and_retrieved_normally(self):
        """VRF keys (starting with _vrf_) are stored like any other CIF key."""
        vrf_key = '_vrf_DIFF003_somecif'
        self.s.save_settings_list('text_templates', vrf_key, ['Response to alert DIFF003.'])
        loaded = self.s.load_settings_list('text_templates', vrf_key)
        self.assertEqual(['Response to alert DIFF003.'], loaded)
        self.assertIn(vrf_key, self.s.list_saved_items('text_templates'))

    def test_multiple_vrf_keys_independent(self):
        self.s.save_settings_list('text_templates', '_vrf_DIFF003_x', ['resp A'])
        self.s.save_settings_list('text_templates', '_vrf_PLAT001_y', ['resp B'])
        self.assertEqual(['resp A'], self.s.load_settings_list('text_templates', '_vrf_DIFF003_x'))
        self.assertEqual(['resp B'], self.s.load_settings_list('text_templates', '_vrf_PLAT001_y'))

    # --- raw export / import round-trip (isolated from AppWindow) ---

    def test_export_import_text_templates_roundtrip(self):
        """Simulate what AppWindow.export_raw_text_templates / import_raw_text_templates do."""
        # Populate a source settings object
        src = FinalCifSettings(settings_path=self.settings_path)
        src.save_settings_list('text_templates', '_refine_special_details',
                               ['All H-atom positions were refined freely.'])
        src.save_settings_list('text_templates', '_exptl_crystal_description',
                               ['Colourless prism', 'Red block'])

        # Export (mimicking AppWindow.export_raw_text_templates)
        templates_list = [
            {'cif_key': key, 'data': src.load_settings_list('text_templates', key)}
            for key in src.list_saved_items('text_templates')
        ]

        # Write to a JSON file (mimicking export_all_templates)
        export_path = self.tmpdir / 'exported.json'
        export_path.write_text(
            json.dumps({'text': templates_list}, ensure_ascii=False, indent=2),
            encoding='utf-8',
        )

        # Re-import into a fresh settings object (mimicking import_raw_text_templates)
        dst_path = self.tmpdir / 'dst_settings.json'
        dst = FinalCifSettings(settings_path=dst_path)
        raw = json.loads(export_path.read_text())
        for entry in raw['text']:
            dst.save_settings_list('text_templates', entry['cif_key'], entry['data'])

        # Verify round-trip fidelity
        self.assertEqual(
            ['All H-atom positions were refined freely.'],
            dst.load_settings_list('text_templates', '_refine_special_details'),
        )
        self.assertEqual(
            ['Colourless prism', 'Red block'],
            dst.load_settings_list('text_templates', '_exptl_crystal_description'),
        )

    def test_import_raw_text_templates_handles_none_gracefully(self):
        """import_raw_text_templates(None) (missing 'text' key) must not raise."""
        # This mimics: self.import_raw_text_templates(templates.get('text'))
        # when the loaded file has no 'text' key.
        templates_list = None
        # Should be a no-op, not crash
        if templates_list:
            for entry in templates_list:
                self.s.save_settings_list('text_templates', entry['cif_key'], entry['data'])
